# pylint: disable=invalid-name
# -*- coding: utf-8 -*-
#
#       ramstk.analyses.milhdbk217f.models.Semiconductor.py is part of the
#       RAMSTK Project
#
# All rights reserved.
# Copyright 2007 - 2020 Doyle Rowland doyle.rowland <AT> reliaqual <DOT> com
"""Semiconductor MIL-HDBK-217F Calculations Module."""

# Standard Library Imports
from math import exp, log, sqrt
from typing import Any, Dict, List

PART_COUNT_LAMBDA_B_DICT = {
    1: {
        1: [
            0.00360,
            0.0280,
            0.049,
            0.043,
            0.100,
            0.092,
            0.210,
            0.200,
            0.44,
            0.170,
            0.00180,
            0.076,
            0.23,
            1.50,
        ],
        2: [
            0.00094,
            0.0075,
            0.013,
            0.011,
            0.027,
            0.024,
            0.054,
            0.054,
            0.12,
            0.045,
            0.00047,
            0.020,
            0.06,
            0.40,
        ],
        3: [
            0.06500,
            0.5200,
            0.890,
            0.780,
            1.900,
            1.700,
            3.700,
            3.700,
            8.00,
            3.100,
            0.03200,
            1.400,
            4.10,
            28.0,
        ],
        4: [
            0.00280,
            0.0220,
            0.039,
            0.034,
            0.062,
            0.073,
            0.160,
            0.160,
            0.35,
            0.130,
            0.00140,
            0.060,
            0.18,
            1.20,
        ],
        5: [
            0.00290,
            0.0230,
            0.040,
            0.035,
            0.084,
            0.075,
            0.170,
            0.170,
            0.36,
            0.140,
            0.00150,
            0.062,
            0.18,
            1.20,
        ],
        6: [
            0.00330,
            0.0240,
            0.039,
            0.035,
            0.082,
            0.066,
            0.150,
            0.130,
            0.27,
            0.120,
            0.00160,
            0.060,
            0.16,
            1.30,
        ],
        7: [
            0.00580,
            0.0400,
            0.066,
            0.060,
            0.140,
            0.110,
            0.250,
            0.220,
            0.460,
            0.21,
            0.00280,
            0.100,
            0.28,
            2.10,
        ],
    },
    2: {
        1: [
            0.86,
            2.80,
            8.9,
            5.6,
            20.0,
            11.0,
            14.0,
            36.0,
            62.0,
            44.0,
            0.43,
            16.0,
            67.0,
            350.0,
        ],
        2: [
            0.31,
            0.76,
            2.1,
            1.5,
            4.60,
            2.00,
            2.50,
            4.50,
            7.60,
            7.90,
            0.16,
            3.70,
            12.0,
            94.00,
        ],
        3: [
            0.004,
            0.0096,
            0.0026,
            0.0019,
            0.058,
            0.025,
            0.032,
            0.057,
            0.097,
            0.10,
            0.002,
            0.048,
            0.15,
            1.2,
        ],
        4: [
            0.028,
            0.068,
            0.19,
            0.14,
            0.41,
            0.18,
            0.22,
            0.40,
            0.69,
            0.71,
            0.014,
            0.34,
            1.1,
            8.5,
        ],
        5: [
            0.047,
            0.11,
            0.31,
            0.23,
            0.68,
            0.3,
            0.37,
            0.67,
            1.1,
            1.2,
            0.023,
            0.56,
            1.8,
            14.0,
        ],
        6: [
            0.0043,
            0.010,
            0.029,
            0.021,
            0.063,
            0.028,
            0.034,
            0.062,
            0.11,
            0.11,
            0.0022,
            0.052,
            0.17,
            1.3,
        ],
    },
    3: {
        1: [
            0.00015,
            0.0011,
            0.0017,
            0.0017,
            0.0037,
            0.0030,
            0.0067,
            0.0060,
            0.013,
            0.0056,
            0.000073,
            0.0027,
            0.0074,
            0.056,
        ],
        2: [
            0.0057,
            0.042,
            0.069,
            0.063,
            0.15,
            0.12,
            0.26,
            0.23,
            0.50,
            0.22,
            0.0029,
            0.11,
            0.29,
            1.1,
        ],
    },
    8: {
        1: [0.17, 0.51, 1.5, 1.0, 3.4, 1.8, 2.3, 5.4, 9.2, 7.2, 0.083, 2.8, 11.0, 63.0],
        2: [
            0.42,
            1.3,
            3.8,
            2.5,
            8.5,
            4.5,
            5.6,
            13.0,
            23.0,
            18.0,
            0.21,
            6.9,
            27.0,
            160.0,
        ],
    },
    11: {
        1: [
            0.01100,
            0.0290,
            0.0830,
            0.0590,
            0.1800,
            0.0840,
            0.1100,
            0.2100,
            0.3500,
            0.3400,
            0.00570,
            0.1500,
            0.510,
            3.70,
        ],
        2: [
            0.02700,
            0.0700,
            0.2000,
            0.1400,
            0.4300,
            0.2000,
            0.2500,
            0.4900,
            0.8300,
            0.8000,
            0.01300,
            0.3500,
            1.200,
            8.70,
        ],
        3: [
            0.00047,
            0.0012,
            0.0035,
            0.0025,
            0.0077,
            0.0035,
            0.0044,
            0.0086,
            0.0150,
            0.0140,
            0.00024,
            0.0053,
            0.021,
            0.15,
        ],
    },
    13: {
        1: [
            5.1,
            16.0,
            49.0,
            32.0,
            110.0,
            58.0,
            72.0,
            100.0,
            170.0,
            230.0,
            2.6,
            87.0,
            350.0,
            2000.0,
        ],
        2: [
            8.9,
            28.0,
            85.0,
            55.0,
            190.0,
            100.0,
            130.0,
            180.0,
            300.0,
            400.0,
            4.5,
            150.0,
            600.0,
            3500.0,
        ],
    },
}
PART_COUNT_LAMBDA_B_LIST = {
    4: [
        0.014,
        0.099,
        0.16,
        0.15,
        0.34,
        0.28,
        0.62,
        0.53,
        1.1,
        0.51,
        0.0069,
        0.25,
        0.68,
        5.3,
    ],
    5: [
        0.016,
        0.12,
        0.20,
        0.18,
        0.42,
        0.35,
        0.80,
        0.74,
        1.6,
        0.66,
        0.0079,
        0.31,
        0.88,
        6.4,
    ],
    6: [0.094, 0.23, 0.63, 0.46, 1.4, 0.60, 0.75, 1.3, 2.3, 2.4, 0.047, 1.1, 3.6, 28.0],
    7: [
        0.074,
        0.15,
        0.37,
        0.29,
        0.81,
        0.29,
        0.37,
        0.52,
        0.88,
        0.037,
        0.33,
        0.66,
        1.8,
        18.0,
    ],
    9: [
        0.014,
        0.099,
        0.16,
        0.15,
        0.34,
        0.28,
        0.62,
        0.53,
        1.1,
        0.51,
        0.0069,
        0.25,
        0.68,
        5.3,
    ],
    10: [
        0.0025,
        0.020,
        0.034,
        0.030,
        0.072,
        0.064,
        0.14,
        0.14,
        0.31,
        0.12,
        0.0012,
        0.053,
        0.16,
        1.1,
    ],
    12: [
        0.0062,
        0.016,
        0.045,
        0.032,
        0.10,
        0.046,
        0.058,
        0.11,
        0.19,
        0.18,
        0.0031,
        0.082,
        0.28,
        2.0,
    ],
}
PART_COUNT_PI_Q: Dict[int, List[float]] = {
    1: [0.7, 1.0, 2.4, 5.5, 8.0],
    3: [0.7, 1.0, 2.4, 5.5, 8.0],
    4: [0.7, 1.0, 2.4, 5.5, 8.0],
    5: [0.7, 1.0, 2.4, 5.5, 8.0],
    6: [0.7, 1.0, 2.4, 5.5, 8.0],
    7: [0.7, 1.0, 2.4, 5.5, 8.0],
    8: [0.7, 1.0, 2.4, 5.5, 8.0],
    9: [0.7, 1.0, 2.4, 5.5, 8.0],
    10: [0.7, 1.0, 2.4, 5.5, 8.0],
    11: [0.7, 1.0, 2.4, 5.5, 8.0],
    12: [0.7, 1.0, 2.4, 5.5, 8.0],
    13: [1.0, 1.0, 3.3],
}
PART_COUNT_PI_Q_HF_DIODE: List[List[float]] = [
    [0.5, 1.0, 5.0, 25, 50],
    [0.5, 1.0, 1.8, 2.5],
]

PART_STRESS_PI_Q: Dict[int, List[float]] = {
    1: [0.7, 1.0, 2.4, 5.5, 8.0],
    3: [0.7, 1.0, 2.4, 5.5, 8.0],
    4: [0.7, 1.0, 2.4, 5.5, 8.0],
    5: [0.7, 1.0, 2.4, 5.5, 8.0],
    6: [0.5, 1.0, 2.0, 5.0],
    7: [0.5, 1.0, 2.0, 5.0],
    8: [0.5, 1.0, 2.0, 5.0],
    9: [0.5, 1.0, 2.0, 5.0],
    10: [0.7, 1.0, 2.4, 5.5, 8.0],
    11: [0.7, 1.0, 2.4, 5.5, 8.0],
    12: [0.7, 1.0, 2.4, 5.5, 8.0],
    13: [1.0, 1.0, 3.3],
}
PART_STRESS_PI_Q_HF_DIODE: Dict[int, List[float]] = {
    1: [0.5, 1.0, 5.0, 25.0, 50.0],
    2: [0.5, 1.0, 5.0, 25.0, 50.0],
    3: [0.5, 1.0, 5.0, 25.0, 50.0],
    4: [0.5, 1.0, 5.0, 25.0, 50.0],
    5: [0.5, 1.0, 1.8, 2.5],
    6: [0.5, 1.0, 5.0, 25.0, 50.0],
}

# Constants used to calculate the temperature factor (piT)
PI_T_DICT: Dict[int, List[float]] = {1: [2903.0, 0.1, 2.0], 2: [5794.0, 0.38, 7.55]}
PI_T_LIST: Dict[int, List[float]] = {
    1: [3091.0, 3091.0, 3091.0, 3091.0, 3091.0, 3091.0, 1925.0, 1925.0],
    2: [5260.0, 2100.0, 2100.0, 2100.0, 2100.0, 2100.0],
}
PI_T_SCALAR: Dict[int, float] = {
    3: 2114.0,
    4: 1925.0,
    5: 2483.0,
    6: 2114.0,
    8: 4485.0,
    9: 1925.0,
    10: 3082.0,
    11: 2790.0,
    12: 2790.0,
    13: 4635.0,
}

# Constants used to calculate the junction temperature.
CASE_TEMPERATURE = [
    35.0,
    45.0,
    50.0,
    45.0,
    50.0,
    60.0,
    60.0,
    75.0,
    75.0,
    60.0,
    35.0,
    50.0,
    60.0,
    45.0,
]
THETA_JC = [
    70.0,
    10.0,
    70.0,
    70.0,
    70.0,
    70.0,
    70.0,
    5.0,
    70.0,
    70.0,
    10.0,
    70.0,
    70.0,
    70.0,
    5.0,
    5.0,
    5.0,
    5.0,
    5.0,
    5.0,
    10.0,
    70.0,
    70.0,
    5.0,
    22.0,
    70.0,
    5.0,
    70.0,
    5.0,
    5.0,
    1.0,
    10.0,
    70.0,
    70.0,
    5.0,
    5.0,
    5.0,
    10.0,
    5.0,
    5.0,
    10.0,
    5.0,
    10.0,
    10.0,
    10.0,
    5.0,
    70.0,
    5.0,
    70.0,
    70.0,
    70.0,
    70.0,
    70.0,
    70.0,
    70.0,
    70.0,
    70.0,
    70.0,
    70.0,
    70.0,
    70.0,
    70.0,
    70.0,
    70.0,
    70.0,
]

# Constants used to calculate the construction factor (piC).
PI_C = [1.0, 2.0]

PI_E = {
    1: [1.0, 6.0, 9.0, 9.0, 19.0, 13.0, 29.0, 20.0, 43.0, 24.0, 0.5, 14.0, 32.0, 320.0],
    2: [1.0, 2.0, 5.0, 4.0, 11.0, 4.0, 5.0, 7.0, 12.0, 16.0, 0.5, 9.0, 24.0, 250.0],
    3: [1.0, 6.0, 9.0, 9.0, 19.0, 13.0, 29.0, 20.0, 43.0, 24.0, 0.5, 14.0, 32.0, 320.0],
    4: [1.0, 6.0, 9.0, 9.0, 19.0, 13.0, 29.0, 20.0, 43.0, 24.0, 0.5, 14.0, 32.0, 320.0],
    5: [1.0, 6.0, 9.0, 9.0, 19.0, 13.0, 29.0, 20.0, 43.0, 24.0, 0.5, 14.0, 32.0, 320.0],
    6: [1.0, 2.0, 5.0, 4.0, 11.0, 4.0, 5.0, 7.0, 12.0, 16.0, 0.5, 9.0, 24.0, 250.0],
    7: [1.0, 2.0, 5.0, 4.0, 11.0, 4.0, 5.0, 7.0, 12.0, 16.0, 0.5, 9.0, 24.0, 250.0],
    8: [1.0, 2.0, 5.0, 4.0, 11.0, 4.0, 5.0, 7.0, 12.0, 16.0, 0.5, 7.5, 24.0, 250.0],
    9: [1.0, 6.0, 9.0, 9.0, 19.0, 13.0, 29.0, 20.0, 43.0, 24.0, 0.5, 14.0, 32.0, 320.0],
    10: [
        1.0,
        6.0,
        9.0,
        9.0,
        19.0,
        13.0,
        29.0,
        20.0,
        43.0,
        24.0,
        0.5,
        14.0,
        32.0,
        320.0,
    ],
    11: [1.0, 2.0, 8.0, 5.0, 12.0, 4.0, 6.0, 6.0, 8.0, 17.0, 0.5, 9.0, 24.0, 450.0],
    12: [1.0, 2.0, 8.0, 5.0, 12.0, 4.0, 6.0, 6.0, 8.0, 17.0, 0.5, 9.0, 24.0, 450.0],
    13: [1.0, 2.0, 8.0, 5.0, 12.0, 4.0, 6.0, 6.0, 8.0, 17.0, 0.5, 9.0, 24.0, 450.0],
}

# Constants used to calculate the matching factor (piM).
PI_M = [1.0, 2.0, 4.0]


def calculate_application_factor(attributes: Dict[str, Any]) -> Dict[str, Any]:
    """Calculate the application factor (piA) for the semiconductor device.

    :param attributes: the attributes of the semiconductor being calculated.
    :return attributes: the updated attributes of the semiconductor being
        calculated.
    :rtype: dict
    :raise: IndexError if passed an unknown application ID.
    """
    _functions = {
        2: _get_section_6_2_application_factor,
        3: _get_section_6_3_application_factor,
        4: _get_section_6_4_application_factor,
        7: _get_section_6_7_application_factor,
        8: _get_section_6_8_application_factor,
        13: _get_section_6_13_application_factor,
    }

    try:
        attributes = _functions[attributes["subcategory_id"]](attributes)
    except KeyError:
        attributes["piA"] = 0.0

    return attributes


def calculate_electrical_stress_factor(attributes: Dict[str, Any]) -> Dict[str, Any]:
    """Calculate the electrical stress factor for the semiconductor device.

    :param attributes: the attributes of the semiconductor being
        calculated.
    :return attributes: the updated attributes of the semiconductor being
        calculated.
    :rtype: dict
    """
    _functions = {
        1: _get_section_6_1_electrical_stress_factor,
        3: _get_section_6_3_electrical_stress_factor,
        6: _get_section_6_6_electrical_stress_factor,
        10: _get_section_6_10_electrical_stress_factor,
    }

    try:
        attributes = _functions[attributes["subcategory_id"]](attributes)
    except KeyError:
        attributes["piS"] = 0.0

    return attributes


def calculate_junction_temperature(attributes: Dict[str, Any]) -> Dict[str, Any]:
    """Calculate the junction temperature of the semiconductor device.

    .. note:: This function will also estimate the case temperature if it is
        passed in at less than or equal to zero.

    .. note:: This function will also estimate the junction-case thermal
        resistance (thetaJC) if it is passed in at less than or equal to zero.

    :param attributes: the attributes of the semiconductor being
        calculated.
    :return attributes: the updated attributes of the semiconductor being
        calculated.
    :rtype: dict
    :raise: IndexError if passed an unknown active environment ID when the case
        temperature is passed at <=0.0 or an unknown package ID when the
        junction-case thermal resistance is passed at <=0.0.
    """
    if attributes["temperature_case"] <= 0.0:
        attributes["temperature_case"] = CASE_TEMPERATURE[
            attributes["environment_active_id"] - 1
        ]

    if attributes["theta_jc"] <= 0.0:
        attributes["theta_jc"] = THETA_JC[attributes["package_id"] - 1]

    attributes["temperature_junction"] = (
        attributes["temperature_case"]
        + attributes["theta_jc"] * attributes["power_operating"]
    )

    return attributes


def calculate_part_count(**attributes: Dict[str, Any]) -> Dict[str, Any]:
    """Calculate the parts count hazard rate for a semiconductor.

    :return: attributes; the keyword argument (hardware attribute) dictionary
        with updated values.
    :rtype: dict
    """
    return get_part_count_lambda_b(attributes)


# pylint: disable=too-many-locals
def calculate_part_stress(**attributes: Dict[str, Any]) -> Dict[str, Any]:
    """Calculate the part stress active hazard rate for a semiconductor.

    This function calculates the MIL-HDBK-217F hazard rate using the part
    stress method.

    :return: attributes; the keyword argument (hardware attribute)
        dictionary with updated values.
    :rtype: dict
    :raise: IndexError if passed an unknown construction ID or matching ID.
    """
    attributes = get_part_stress_quality_factor(attributes)
    attributes = calculate_part_stress_lambda_b(attributes)
    attributes = calculate_junction_temperature(attributes)
    attributes = calculate_temperature_factor(attributes)
    attributes = calculate_application_factor(attributes)
    attributes = calculate_power_rating_factor(attributes)
    attributes = calculate_electrical_stress_factor(attributes)

    # Retrieve the construction factor (piC).
    attributes["piC"] = PI_C[attributes["construction_id"] - 1]  # type: ignore

    # Retrieve the matching network factor (piM).
    attributes["piM"] = PI_M[attributes["matching_id"] - 1]  # type: ignore

    # Calculate forward current factor (piI) and power degradation factor (piP)
    attributes["piI"] = attributes["current_operating"] ** 0.68  # type: ignore
    attributes["piP"] = 1.0 / (  # type: ignore
        2.0 * (1.0 - attributes["power_ratio"])
    )  # type: ignore

    attributes["hazard_rate_active"] = (
        attributes["lambda_b"]  # type: ignore
        * attributes["piT"]
        * attributes["piQ"]
        * attributes["piE"]
    )

    if attributes["subcategory_id"] == 1:
        attributes["hazard_rate_active"] = (
            attributes["hazard_rate_active"]  # type: ignore
            * attributes["piS"]
            * attributes["piC"]
        )
    elif attributes["subcategory_id"] == 2:
        attributes["hazard_rate_active"] = (
            attributes["hazard_rate_active"]  # type: ignore
            * attributes["piA"]
            * attributes["piR"]
        )
    elif attributes["subcategory_id"] == 3:
        attributes["hazard_rate_active"] = (
            attributes["hazard_rate_active"]  # type: ignore
            * attributes["piA"]
            * attributes["piR"]
            * attributes["piS"]
        )
    elif attributes["subcategory_id"] == 4:
        attributes["hazard_rate_active"] = (
            attributes["hazard_rate_active"] * attributes["piA"]  # type: ignore
        )
    elif attributes["subcategory_id"] in [6, 10]:
        attributes["hazard_rate_active"] = (
            attributes["hazard_rate_active"]  # type: ignore
            * attributes["piR"]
            * attributes["piS"]
        )
    elif attributes["subcategory_id"] in [7, 8]:
        attributes["hazard_rate_active"] = (
            attributes["hazard_rate_active"]  # type: ignore
            * attributes["piA"]
            * attributes["piM"]
        )
    elif attributes["subcategory_id"] == 13:
        attributes["hazard_rate_active"] = (
            attributes["hazard_rate_active"]  # type: ignore
            * attributes["piI"]
            * attributes["piA"]
            * attributes["piP"]
        )

    return attributes


def calculate_part_stress_lambda_b(attributes: Dict[str, Any]) -> Dict[str, Any]:
    """Retrieve MIL-HDBK-217F base hazard rate for the semiconductor device.

    :param attributes: the attributes of the semiconductor being
        calculated.
    :return attributes: the updated attributes of the semiconductor being
        calculated.
    :rtype: dict
    :raise: IndexError if passed an unknown type ID.
    :raise: KeyError if passed an unkown subcategory ID.
    """
    _dic_lambdab_scalar: Dict[int, float] = {3: 0.00074, 5: 0.0083, 6: 0.18, 10: 0.0022}
    _dic_lambdab_list: Dict[int, List[float]] = {
        1: [0.0038, 0.0010, 0.069, 0.003, 0.005, 0.0013, 0.0034, 0.002],
        2: [0.22, 0.18, 0.0023, 0.0081, 0.027, 0.0025, 0.0025],
        4: [0.012, 0.0045],
        9: [0.06, 0.023],
        11: [
            0.0055,
            0.004,
            0.0025,
            0.013,
            0.013,
            0.0064,
            0.0033,
            0.017,
            0.017,
            0.0086,
            0.0013,
            0.00023,
        ],
        13: [3.23, 5.65],
    }

    if attributes["subcategory_id"] in [3, 5, 6, 10]:
        attributes["lambda_b"] = _dic_lambdab_scalar[attributes["subcategory_id"]]
    elif attributes["subcategory_id"] == 7:
        attributes["lambda_b"] = 0.032 * exp(
            0.354 * attributes["frequency_operating"]
            + 0.00558 * attributes["power_operating"]
        )
    elif attributes["subcategory_id"] == 8:
        if (
            1.0 < attributes["frequency_operating"] <= 10.0
            and attributes["power_operating"] < 0.1
        ):
            attributes["lambda_b"] = 0.052
        else:
            attributes["lambda_b"] = 0.0093 * exp(
                0.429 * attributes["frequency_operating"]
                + 0.486 * attributes["power_operating"]
            )
    elif attributes["subcategory_id"] == 12:
        if attributes["application_id"] in [1, 3]:
            attributes["lambda_b"] = 0.00043 * attributes["n_elements"] + 0.000043
        else:
            attributes["lambda_b"] = 0.00043 * attributes["n_elements"]
    else:
        attributes["lambda_b"] = _dic_lambdab_list[attributes["subcategory_id"]][
            attributes["type_id"] - 1
        ]

    return attributes


def calculate_power_rating_factor(attributes: Dict[str, Any]) -> Dict[str, Any]:
    """Calculate the power rating factor for the semiconductor device.

    :param attributes: the attributes of the semiconductor being
        calculated.
    :return attributes: the updated attributes of the semiconductor being
        calculated.
    :rtype: dict
    :raise: TypeError if passed a string for the rated power or rated current.
    :raise: ValueError if passed a rated power <=0.0.
    """
    if attributes["subcategory_id"] == 2:
        attributes = _get_section_6_2_power_rating_factor(attributes)
    elif attributes["subcategory_id"] == 3:
        attributes = _get_section_6_3_power_rating_factor(attributes)
    elif attributes["subcategory_id"] == 6:
        attributes = _get_section_6_6_power_rating_factor(attributes)
    elif attributes["subcategory_id"] == 10:
        attributes["piR"] = attributes["current_rated"] ** 0.4
    else:
        attributes["piR"] = 0.0

    return attributes


def calculate_temperature_factor(attributes: Dict[str, Any]) -> Dict[str, Any]:
    """Calculate the temperature factor for the semiconductor device.

    :param attributes: the attributes of the semiconductor being
        calculated.
    :return attributes: the updated attributes of the semiconductor being
        calculated.
    :rtype: dict
    :raise: IndexError if passed an unknown type ID.
    :raise: KeyError if passed an unknown subcategory ID.
    """
    if attributes["subcategory_id"] in [1, 2]:
        _factors = PI_T_LIST[attributes["subcategory_id"]][attributes["type_id"] - 1]
    elif attributes["subcategory_id"] == 7:
        _factors = PI_T_DICT[attributes["type_id"]]
    else:
        _factors = PI_T_SCALAR[attributes["subcategory_id"]]

    if attributes["subcategory_id"] == 7:
        _f0 = _factors[0]
        _f1 = _factors[1]
        _f2 = _factors[2]
        if attributes["voltage_ratio"] <= 0.4:
            attributes["piT"] = _f1 * exp(
                -_f0
                * (1.0 / (attributes["temperature_junction"] + 273.0) - 1.0 / 298.0)
            )
        else:
            attributes["piT"] = (
                _f2
                * (attributes["voltage_ratio"] - 0.35)
                * exp(
                    -_f0
                    * (1.0 / (attributes["temperature_junction"] + 273.0) - 1.0 / 298.0)
                )
            )
    else:
        attributes["piT"] = exp(
            -_factors
            * (1.0 / (attributes["temperature_junction"] + 273.0) - 1.0 / 298.0)
        )

    return attributes


def get_part_count_lambda_b(attributes: Dict[str, Any]) -> Dict[str, Any]:
    """Retrieve MIL-HDBK-217F base hazard rate for the semiconductor device.

    This function retrieves the MIL-HDBK-217F hazard rate from the dictionary
    PART_COUNT_LAMBDA_B.  Keys are for PART_COUNT_LAMBDA_B are:

        #. subcategory_id
        #. environment_active_id
        #. type id; if the semiconductor subcategory is NOT type dependent,
            then the second key will be zero.

    Current subcategory IDs are:

    +----------------+-------------------------------+-----------------+
    | Subcategory    |          Semiconductor        | MIL-HDBK-217F   |
    |       ID       |              Style            |    Section      |
    +================+===============================+=================+
    |        1       | Diode, Low Frequency          |        6.1      |
    +----------------+-------------------------------+-----------------+
    |        2       | Diode, High Frequency         |        6.2      |
    +----------------+-------------------------------+-----------------+
    |        3       | Transistor, Low Frequency,    |        6.3      |
    |                | Bipolar                       |                 |
    +----------------+-------------------------------+-----------------+
    |        4       | Transistor, Low Frequency,    |        6.4      |
    |                | Si FET                        |                 |
    +----------------+-------------------------------+-----------------+
    |        5       | Transistor, Unijunction       |        6.5      |
    +----------------+-------------------------------+-----------------+
    |        6       | Transistor, High Frequency,   |        6.6      |
    |                | Low Noise,Bipolar             |                 |
    +----------------+-------------------------------+-----------------+
    |        7       | Transistor, High Frequency,   |        6.7      |
    |                | High Power, Bipolar           |                 |
    +----------------+-------------------------------+-----------------+
    |        8       | Transistor, High Frequency,   |        6.8      |
    |                | GaAs FET                      |                 |
    +----------------+-------------------------------+-----------------+
    |        9       | Transistor, High Frequency,   |        6.9      |
    |                | Si FET                        |                 |
    +----------------+-------------------------------+-----------------+
    |       10       | Thyristor/SCR                 |       6.10      |
    +----------------+-------------------------------+-----------------+
    |       11       | Optoelectronic, Detector,     |       6.11      |
    |                | Isolator, Emitter             |                 |
    +----------------+-------------------------------+-----------------+
    |       12       | Optoelectronic, Alphanumeric  |       6.12      |
    |                | Display                       |                 |
    +----------------+-------------------------------+-----------------+
    |       13       | Optoelectronic, Laser Diode   |       6.13      |
    +----------------+-------------------------------+-----------------+

    :param attributes: the attributes of the semiconductor being
        calculated.
    :return: _base_hr; the parts count base hazard rate.
    :rtype: float
    :raise: IndexError if passed an unknown active environment ID.
    :raise: KeyError if passed an unknown subcategory ID or type ID.
    """
    if attributes["subcategory_id"] in [1, 2, 3, 8, 11, 13]:
        attributes["lambda_b"] = PART_COUNT_LAMBDA_B_DICT[attributes["subcategory_id"]][
            attributes["type_id"]
        ][attributes["environment_active_id"] - 1]
    else:
        attributes["lambda_b"] = PART_COUNT_LAMBDA_B_LIST[attributes["subcategory_id"]][
            attributes["environment_active_id"] - 1
        ]

    return attributes


def get_part_count_quality_factor(attributes: Dict[str, Any]) -> Dict[str, Any]:
    """Retrieve the parts count quality factor for the semiconductor device.

    :param attributes: the attributes of the semiconductor being
        calculated.
    :return attributes: the updated attributes of the semiconductor being
        calculated.
    :rtype: dict
    :raise: IndexError if passed an unknown quality ID.
    :raise: KeyError if passed an unknown subcategory ID.
    """
    if attributes["subcategory_id"] == 2:
        if attributes["type_id"] == 5:
            attributes["piQ"] = PART_COUNT_PI_Q_HF_DIODE[1][
                attributes["quality_id"] - 1
            ]
        else:
            attributes["piQ"] = PART_COUNT_PI_Q_HF_DIODE[0][
                attributes["quality_id"] - 1
            ]
    else:
        attributes["piQ"] = PART_COUNT_PI_Q[attributes["subcategory_id"]][
            attributes["quality_id"] - 1
        ]

    return attributes


def get_part_stress_quality_factor(attributes: Dict[str, Any]) -> Dict[str, Any]:
    """Select the part stress quality factor for the semiconductor device.

    :param attributes: the attributes of the semiconductor being
        calculated.
    :return attributes: the updated attributes of the semiconductor being
        calculated.
    :rtype: dict
    :raise: IndexError if passed an unknown quality ID.
    :raise: KeyError if passed an unknown subcategory ID.
    """
    if attributes["subcategory_id"] == 2:
        attributes["piQ"] = PART_STRESS_PI_Q_HF_DIODE[attributes["type_id"]][
            attributes["quality_id"] - 1
        ]
    else:
        attributes["piQ"] = PART_STRESS_PI_Q[attributes["subcategory_id"]][
            attributes["quality_id"] - 1
        ]

    return attributes


def _get_section_6_1_electrical_stress_factor(
    attributes: Dict[str, Any]
) -> Dict[str, Any]:
    """Get piS and set default values for LF diodes.

    This function is for MIL-HDBK-217F, Section 6.2 devices.  The
    voltage stress ratio default value is also set.

    :param attributes: the attributes of the semiconductor being
        calculated.
    :return attributes: the updated attributes of the semiconductor being
        calculated.
    :rtype: dict
    """
    if attributes["type_id"] < 5:
        attributes["voltage_ratio"] = attributes["voltage_ratio"] or 0.7
        if attributes["voltage_ratio"] <= 0.3:
            attributes["piS"] = 0.054
        else:
            attributes["piS"] = attributes["voltage_ratio"] ** 2.43
    else:
        attributes["piS"] = 1.0

    return attributes


def _get_section_6_2_application_factor(attributes: Dict[str, Any]) -> Dict[str, Any]:
    """Get piA and set default values for HF diodes.

    This function is for MIL-HDBK-217F, Section 6.2 devices.  The
    application ID default value is also set.

    :param attributes: the attributes of the semiconductor being
        calculated.
    :return attributes: the updated attributes of the semiconductor being
        calculated.
    :rtype: dict
    """
    attributes["application_id"] = attributes["application_id"] or 2

    if attributes["type_id"] != 6:
        attributes["application_id"] = attributes["application_id"] or 3

    attributes["piA"] = [
        0.5,
        2.5,
        1.0,
    ][attributes["application_id"] - 1]

    return attributes


def _get_section_6_2_power_rating_factor(attributes: Dict[str, Any]) -> Dict[str, Any]:
    """Get piR and set default values for HF diodes.

    This function is for MIL-HDBK-217F, Section 6.3 devices.  The
    rated power default value is also set.

    :param attributes: the attributes of the semiconductor being
        calculated.
    :return attributes: the updated attributes of the semiconductor being
        calculated.
    :rtype: dict
    """
    if attributes["type_id"] == 4:
        attributes["power_rated"] = attributes["power_rated"] or 1000.0
        attributes["piR"] = 0.326 * log(attributes["power_rated"]) - 0.25
    else:
        attributes["piR"] = 1.0

    return attributes


def _get_section_6_3_application_factor(attributes: Dict[str, Any]) -> Dict[str, Any]:
    """Get piA and set default values for LF, LP BJT.

    This function is for MIL-HDBK-217F, Section 6.3 devices.  The
    application ID default value is also set.

    :param attributes: the attributes of the semiconductor being
        calculated.
    :return attributes: the updated attributes of the semiconductor being
        calculated.
    :rtype: dict
    """
    attributes["application_id"] = attributes["application_id"] or {
        0: 0,
        1: 2,
        2: 1,
    }[attributes["type_id"]]

    attributes["piA"] = [
        1.5,
        0.7,
    ][attributes["application_id"] - 1]

    return attributes


def _get_section_6_3_electrical_stress_factor(
    attributes: Dict[str, Any]
) -> Dict[str, Any]:
    """Get piS and set default values for LF, LP BJT.

    This function is for MIL-HDBK-217F, Section 6.3 devices.  The
    voltage stress ratio default value is also set.

    :param attributes: the attributes of the semiconductor being
        calculated.
    :return attributes: the updated attributes of the semiconductor being
        calculated.
    :rtype: dict
    """
    attributes["voltage_ratio"] = attributes["voltage_ratio"] or {
        0: 0.0,
        1: 0.5,
        2: 0.8,
    }[attributes["type_id"]]

    attributes["piS"] = 0.045 * exp(3.1 * attributes["voltage_ratio"])

    return attributes


def _get_section_6_3_power_rating_factor(attributes: Dict[str, Any]) -> Dict[str, Any]:
    """Get piR and set default values for LF, LP BJT.

    This function is for MIL-HDBK-217F, Section 6.3 devices.  The rated power
    default value is also set.

    :param attributes: the attributes of the semiconductor being
        calculated.
    :return attributes: the updated attributes of the semiconductor being
        calculated.
    :rtype: dict
    """
    attributes["power_rated"] = attributes["power_rated"] or {
        0: 0.0,
        1: 0.5,
        2: 100.0,
    }[attributes["type_id"]]

    if attributes["power_rated"] < 0.1:
        attributes["piR"] = 0.43
    else:
        attributes["piR"] = attributes["power_rated"] ** 0.37

    return attributes


def _get_section_6_4_application_factor(attributes: Dict[str, Any]) -> Dict[str, Any]:
    """Get piA and set default values for LF Si FET.

    This function is for MIL-HDBK-217F, Section 6.4 devices.  The
    application ID default value is also set.

    :param attributes: the attributes of the semiconductor being
        calculated.
    :return attributes: the updated attributes of the semiconductor being
        calculated.
    :rtype: dict
    """
    attributes["application_id"] = attributes["application_id"] or 2

    attributes["piA"] = [
        1.5,
        0.7,
        2.0,
        4.0,
        8.0,
        10.0,
    ][attributes["application_id"] - 1]

    return attributes


def _get_section_6_6_electrical_stress_factor(
    attributes: Dict[str, Any]
) -> Dict[str, Any]:
    """Get piS and set default values for HF, LN BJT.

    This function is for MIL-HDBK-217F, Section 6.6 devices.  The
    voltage stress ratio default value is also set.

    :param attributes: the attributes of the semiconductor being
        calculated.
    :return attributes: the updated attributes of the semiconductor being
        calculated.
    :rtype: dict
    """
    attributes["voltage_ratio"] = attributes["voltage_ratio"] or 0.7

    attributes["piS"] = 0.045 * exp(3.1 * attributes["voltage_ratio"])

    return attributes


def _get_section_6_6_power_rating_factor(attributes: Dict[str, Any]) -> Dict[str, Any]:
    """Get piR and set default values for HF, LN BJT.

    This function is for MIL-HDBK-217F, Section 6.6 devices.  The rated power
    default value is also set.

    :param attributes: the attributes of the semiconductor being
        calculated.
    :return attributes: the updated attributes of the semiconductor being
        calculated.
    :rtype: dict
    """
    attributes["power_rated"] = attributes["power_rated"] or 0.5

    if attributes["power_rated"] < 0.1:
        attributes["piR"] = 0.43
    else:
        attributes["piR"] = attributes["power_rated"] ** 0.37

    return attributes


def _get_section_6_7_application_factor(attributes: Dict[str, Any]) -> Dict[str, Any]:
    """Get piA and set default values for HF and HP BJT devices.

    This function is for MIL-HDBK-217F, Section 6.7 devices.  The
    application ID and the duty cycle default values are also set.

    :param attributes: the attributes of the semiconductor being
        calculated.
    :return attributes: the updated attributes of the semiconductor being
        calculated.
    :rtype: dict
    """
    attributes["application_id"] = attributes["application_id"] or 2
    attributes["duty_cycle"] = attributes["duty_cycle"] or 0.2

    if attributes["application_id"] == 1:
        attributes["piA"] = 7.6
    else:
        attributes["piA"] = 0.06 * (attributes["duty_cycle"] / 100.0) + 0.4

    return attributes


def _get_section_6_8_application_factor(attributes: Dict[str, Any]) -> Dict[str, Any]:
    """Get piA and set default values for GaAs FET.

    This function is for MIL-HDBK-217F, Section 6.8 devices.  The
    application ID default value is also set.

    :param attributes: the attributes of the semiconductor being
        calculated.
    :return attributes: the updated attributes of the semiconductor being
        calculated.
    :rtype: dict
    """
    attributes["application_id"] = attributes["application_id"] or 1

    attributes["piA"] = [
        1.0,
        4.0,
    ][attributes["application_id"] - 1]

    return attributes


def _get_section_6_10_electrical_stress_factor(
    attributes: Dict[str, Any]
) -> Dict[str, Any]:
    """Get piS and set default values for thyristors and SCR.

    This function is for MIL-HDBK-217F, Section 6.10 devices.  The
    rated current and voltage stress ratio default value is also set.

    :param attributes: the attributes of the semiconductor being
        calculated.
    :return attributes: the updated attributes of the semiconductor being
        calculated.
    :rtype: dict
    """
    attributes["current_rated"] = attributes["current_rated"] or 1.0
    attributes["voltage_ratio"] = attributes["voltage_ratio"] or 0.7

    if attributes["voltage_ratio"] <= 0.3:
        attributes["piS"] = 0.1
    else:
        attributes["piS"] = attributes["voltage_ratio"] ** 1.9

    return attributes


def _get_section_6_13_application_factor(attributes: Dict[str, Any]) -> Dict[str, Any]:
    """Get piA and set default values for laser diodes.

    This function is for MIL-HDBK-217F, Section 6.13 devices.  The
    application ID and the duty cycle default values are also set.

    :param attributes: the attributes of the semiconductor being
        calculated.
    :return attributes: the updated attributes of the semiconductor being
        calculated.
    :rtype: dict
    """
    attributes["application_id"] = attributes["application_id"] or 2
    attributes["duty_cycle"] = attributes["duty_cycle"] or 0.6

    if attributes["application_id"] == 1:
        attributes["piA"] = 4.4
    else:
        attributes["piA"] = sqrt(attributes["duty_cycle"] / 100.0)

    return attributes



# Standard Library Imports
from typing import Any

# RAMSTK Package Imports
from ramstk.configuration import RAMSTKUserConfiguration as RAMSTKUserConfiguration
from ramstk.db.base import BaseDatabase as BaseDatabase
from ramstk.views.gtk3 import Gtk as Gtk
from ramstk.views.gtk3 import _ as _
from ramstk.views.gtk3.widgets.dialog import (
    RAMSTKDatabaseSelect as RAMSTKDatabaseSelect,
)
from ramstk.views.gtk3.widgets.dialog import RAMSTKDialog as RAMSTKDialog
from ramstk.views.gtk3.widgets.dialog import RAMSTKMessageDialog as RAMSTKMessageDialog
from ramstk.views.gtk3.widgets.label import RAMSTKLabel as RAMSTKLabel

class CreateProject:
    RAMSTK_USER_CONFIGURATION: RAMSTKUserConfiguration = ...
    _parent: Any = ...
    def __init__(
        self,
        __button: Gtk.ToolButton,
        configuration: RAMSTKUserConfiguration,
        parent: object,
    ) -> None: ...
    def _do_confirm_overwrite(self, database: str) -> None: ...
    def _do_request_create_project(self) -> None: ...

class OpenProject:
    RAMSTK_USER_CONFIGURATION: RAMSTKUserConfiguration = ...
    _parent: Any = ...
    def __init__(
        self,
        __button: Gtk.ToolButton,
        configuration: RAMSTKUserConfiguration,
        parent: object,
    ) -> None: ...
    def _do_request_open_project(self) -> None: ...
    def __project_is_open(self) -> None: ...



# Standard Library Imports
from typing import Any

# RAMSTK Package Imports
from ramstk.db import RAMSTK_BASE as RAMSTK_BASE
from ramstk.models import RAMSTKBaseRecord as RAMSTKBaseRecord

class RAMSTKNSWCRecord(RAMSTK_BASE, RAMSTKBaseRecord):
    __defaults__: Any
    __tablename__: str
    __table_args__: Any
    revision_id: Any
    hardware_id: Any
    Cac: Any
    Calt: Any
    Cb: Any
    Cbl: Any
    Cbt: Any
    Cbv: Any
    Cc: Any
    Ccf: Any
    Ccp: Any
    Ccs: Any
    Ccv: Any
    Ccw: Any
    Cd: Any
    Cdc: Any
    Cdl: Any
    Cdp: Any
    Cds: Any
    Cdt: Any
    Cdw: Any
    Cdy: Any
    Ce: Any
    Cf: Any
    Cg: Any
    Cga: Any
    Cgl: Any
    Cgp: Any
    Cgs: Any
    Cgt: Any
    Cgv: Any
    Ch: Any
    Ci: Any
    Ck: Any
    Cl: Any
    Clc: Any
    Cm: Any
    Cmu: Any
    Cn: Any
    Cnp: Any
    Cnw: Any
    Cp: Any
    Cpd: Any
    Cpf: Any
    Cpv: Any
    Cq: Any
    Cr: Any
    Crd: Any
    Cs: Any
    Csc: Any
    Csf: Any
    Cst: Any
    Csv: Any
    Csw: Any
    Csz: Any
    Ct: Any
    Cv: Any
    Cw: Any
    Cy: Any
    def get_attributes(self): ...



# Standard Library Imports
from typing import Any, Dict, List

# RAMSTK Package Imports
from ramstk.views.gtk3 import Gtk as Gtk
from ramstk.views.gtk3 import _ as _
from ramstk.views.gtk3.widgets import RAMSTKCheckButton as RAMSTKCheckButton
from ramstk.views.gtk3.widgets import RAMSTKComboBox as RAMSTKComboBox
from ramstk.views.gtk3.widgets import RAMSTKEntry as RAMSTKEntry
from ramstk.views.gtk3.widgets import RAMSTKFixedPanel as RAMSTKFixedPanel
from ramstk.views.gtk3.widgets import RAMSTKPlot as RAMSTKPlot
from ramstk.views.gtk3.widgets import RAMSTKScrolledWindow as RAMSTKScrolledWindow
from ramstk.views.gtk3.widgets import RAMSTKTextView as RAMSTKTextView

class DesignElectricEnvironmentalInputPanel(RAMSTKFixedPanel):
    _select_msg: str
    _tag: str
    _title: str
    cmbActiveEnviron: RAMSTKComboBox
    cmbDormantEnviron: RAMSTKComboBox
    txtActiveTemp: RAMSTKEntry
    txtDormantTemp: RAMSTKEntry
    txtDutyCycle: RAMSTKEntry
    txtMissionTime: RAMSTKEntry
    dic_attribute_widget_map: Dict[str, List[Any]]
    def __init__(self) -> None: ...
    def do_load_environment_active(self, environments: List[str]) -> None: ...
    def do_load_environment_dormant(self, environments: List[str]) -> None: ...

class DesignElectricStressInputPanel(RAMSTKFixedPanel):
    _select_msg: str
    _tag: str
    _title: str
    txtTemperatureRatedMin: RAMSTKEntry
    txtTemperatureKnee: RAMSTKEntry
    txtTemperatureRatedMax: RAMSTKEntry
    txtCurrentRated: RAMSTKEntry
    txtCurrentOperating: RAMSTKEntry
    txtPowerRated: RAMSTKEntry
    txtPowerOperating: RAMSTKEntry
    txtVoltageRated: RAMSTKEntry
    txtVoltageAC: RAMSTKEntry
    txtVoltageDC: RAMSTKEntry
    dic_attribute_widget_map: Dict[str, List[Any]]
    def __init__(self) -> None: ...

class DesignElectricStressResultPanel(RAMSTKFixedPanel):
    _select_msg: str
    _tag: str
    _title: str
    chkOverstress: RAMSTKCheckButton
    pltPlot: RAMSTKPlot
    txtCurrentRatio: RAMSTKEntry
    txtPowerRatio: RAMSTKEntry
    txtVoltageRatio: RAMSTKEntry
    txtReason: RAMSTKTextView
    _lst_derate_criteria: Any
    _category_id: int
    _part_number: str
    _ref_des: str
    dic_attribute_widget_map: Dict[str, List[Any]]
    def __init__(self) -> None: ...
    def _do_load_derating_curve(
        self, attributes: Dict[str, Any], stress: str = ...
    ) -> None: ...
    def _do_load_entries(self, attributes: Dict[str, Any]) -> None: ...
    def _do_set_hardware_attributes(self, attributes: Dict[str, Any]) -> None: ...
    def __make_ui(self) -> None: ...



# Standard Library Imports
from typing import Any, Dict, List, Tuple, Union

# Third Party Imports
import matplotlib

# RAMSTK Package Imports
from ramstk.views.gtk3 import Gdk as Gdk
from ramstk.views.gtk3 import Gtk as Gtk
from ramstk.views.gtk3 import _ as _

class RAMSTKPlot:
    _lst_max: Any = ...
    _lst_min: Any = ...
    figure: Any = ...
    canvas: Any = ...
    axis: Any = ...
    plot: Any = ...
    def __init__(self) -> None: ...
    def do_load_plot(
        self, x_values: List[float], y_values: List[float], **kwargs: Dict[str, str]
    ) -> None: ...
    def do_add_line(
        self,
        x_values: List[float],
        y_values: List[float] = ...,
        color: str = ...,
        marker: str = ...,
    ) -> None: ...
    def do_close_plot(
        self, __window: Gtk.Window, __event: Gdk.Event, parent: Gtk.Widget
    ) -> None: ...
    def do_expand_plot(self, event: matplotlib.backend_bases.MouseEvent) -> None: ...
    def do_make_labels(self, label: str, **kwargs: Any) -> matplotlib.text.Text: ...
    def do_make_legend(self, text: Union[Any], **kwargs: Any) -> None: ...
    def do_make_title(
        self, title: str, fontsize: int = ..., fontweight: str = ...
    ) -> matplotlib.text.Text: ...
    def _do_make_date_plot(
        self, x_values: List[float], y_values: List[float], **kwargs: Dict[str, str]
    ) -> None: ...
    def _do_make_histogram(
        self, x_values: List[float], y_values: List[float], **kwargs: Dict[str, str]
    ) -> None: ...
    def _do_make_scatter_plot(
        self, x_values: List[float], y_values: List[float], **kwargs: Dict[str, str]
    ) -> None: ...
    def _do_make_step_plot(
        self, x_values: List[float], y_values: List[float], **kwargs: Dict[str, str]
    ) -> None: ...
    def _get_minimax_ordinates(self) -> Tuple[float, float]: ...



# Standard Library Imports
from typing import Any, Dict, List

# RAMSTK Package Imports
from ramstk.configuration import (
    RAMSTK_ACTIVE_ENVIRONMENTS as RAMSTK_ACTIVE_ENVIRONMENTS,
)
from ramstk.configuration import (
    RAMSTK_DORMANT_ENVIRONMENTS as RAMSTK_DORMANT_ENVIRONMENTS,
)
from ramstk.configuration import RAMSTK_HR_DISTRIBUTIONS as RAMSTK_HR_DISTRIBUTIONS
from ramstk.configuration import RAMSTK_HR_MODELS as RAMSTK_HR_MODELS
from ramstk.configuration import RAMSTK_HR_TYPES as RAMSTK_HR_TYPES
from ramstk.configuration import RAMSTKUserConfiguration as RAMSTKUserConfiguration
from ramstk.logger import RAMSTKLogManager as RAMSTKLogManager
from ramstk.views.gtk3 import Gtk as Gtk
from ramstk.views.gtk3 import _ as _
from ramstk.views.gtk3.design_electric import (
    CapacitorDesignElectricInputPanel as CapacitorDesignElectricInputPanel,
)
from ramstk.views.gtk3.design_electric import (
    ConnectionDesignElectricInputPanel as ConnectionDesignElectricInputPanel,
)
from ramstk.views.gtk3.design_electric import (
    DesignElectricEnvironmentalInputPanel as DesignElectricEnvironmentalInputPanel,
)
from ramstk.views.gtk3.design_electric import (
    DesignElectricStressInputPanel as DesignElectricStressInputPanel,
)
from ramstk.views.gtk3.design_electric import (
    DesignElectricStressResultPanel as DesignElectricStressResultPanel,
)
from ramstk.views.gtk3.design_electric import (
    ICDesignElectricInputPanel as ICDesignElectricInputPanel,
)
from ramstk.views.gtk3.design_electric import (
    InductorDesignElectricInputPanel as InductorDesignElectricInputPanel,
)
from ramstk.views.gtk3.design_electric import (
    MeterDesignElectricInputPanel as MeterDesignElectricInputPanel,
)
from ramstk.views.gtk3.design_electric import (
    MiscDesignElectricInputPanel as MiscDesignElectricInputPanel,
)
from ramstk.views.gtk3.design_electric import (
    RelayDesignElectricInputPanel as RelayDesignElectricInputPanel,
)
from ramstk.views.gtk3.design_electric import (
    ResistorDesignElectricInputPanel as ResistorDesignElectricInputPanel,
)
from ramstk.views.gtk3.design_electric import (
    SemiconductorDesignElectricInputPanel as SemiconductorDesignElectricInputPanel,
)
from ramstk.views.gtk3.design_electric import (
    SwitchDesignElectricInputPanel as SwitchDesignElectricInputPanel,
)
from ramstk.views.gtk3.milhdbk217f import (
    CapacitorMilHdbk217FResultPanel as CapacitorMilHdbk217FResultPanel,
)
from ramstk.views.gtk3.milhdbk217f import (
    ConnectionMilHdbk217FResultPanel as ConnectionMilHdbk217FResultPanel,
)
from ramstk.views.gtk3.milhdbk217f import (
    ICMilHdbk217FResultPanel as ICMilHdbk217FResultPanel,
)
from ramstk.views.gtk3.milhdbk217f import (
    InductorMilHdbk217FResultPanel as InductorMilHdbk217FResultPanel,
)
from ramstk.views.gtk3.milhdbk217f import (
    MeterMilHdbk217FResultPanel as MeterMilHdbk217FResultPanel,
)
from ramstk.views.gtk3.milhdbk217f import (
    MiscellaneousMilHdbk217FResultPanel as MiscellaneousMilHdbk217FResultPanel,
)
from ramstk.views.gtk3.milhdbk217f import (
    RelayMilHdbk217FResultPanel as RelayMilHdbk217FResultPanel,
)
from ramstk.views.gtk3.milhdbk217f import (
    ResistorMilHdbk217FResultPanel as ResistorMilHdbk217FResultPanel,
)
from ramstk.views.gtk3.milhdbk217f import (
    SemiconductorMilHdbk217FResultPanel as SemiconductorMilHdbk217FResultPanel,
)
from ramstk.views.gtk3.milhdbk217f import (
    SwitchMilHdbk217FResultPanel as SwitchMilHdbk217FResultPanel,
)
from ramstk.views.gtk3.reliability import (
    AvailabilityResultsPanel as AvailabilityResultsPanel,
)
from ramstk.views.gtk3.reliability import ReliabilityInputPanel as ReliabilityInputPanel
from ramstk.views.gtk3.reliability import (
    ReliabilityResultsPanel as ReliabilityResultsPanel,
)
from ramstk.views.gtk3.widgets import RAMSTKModuleView as RAMSTKModuleView
from ramstk.views.gtk3.widgets import RAMSTKPanel as RAMSTKPanel
from ramstk.views.gtk3.widgets import RAMSTKWorkView as RAMSTKWorkView

# RAMSTK Local Imports
from . import HardwareGeneralDataPanel as HardwareGeneralDataPanel
from . import HardwareLogisticsPanel as HardwareLogisticsPanel
from . import HardwareMiscellaneousPanel as HardwareMiscellaneousPanel
from . import HardwareTreePanel as HardwareTreePanel

class HardwareModuleView(RAMSTKModuleView):
    _tag: str
    _tablabel: str
    _tabtooltip: str
    _lst_mnu_labels: Any
    _lst_tooltips: Any
    _pnlPanel: Any
    def __init__(
        self, configuration: RAMSTKUserConfiguration, logger: RAMSTKLogManager
    ) -> None: ...
    def _do_request_calculate_hardware(self, __button: Gtk.ToolButton) -> None: ...
    def _do_request_calculate_all_hardware(self, __button: Gtk.ToolButton) -> None: ...
    def _do_request_insert_child(self, __button: Gtk.ToolButton) -> Any: ...
    def _do_request_insert_part(self, __button: Gtk.ToolButton) -> None: ...
    def _do_request_insert_sibling(self, __button: Gtk.ToolButton) -> Any: ...
    _record_id: Any
    _parent_id: Any
    def _do_set_record_id(self, attributes: Dict[str, Any]) -> None: ...
    def __make_ui(self) -> None: ...

class HardwareGeneralDataView(RAMSTKWorkView):
    _tag: str
    _tablabel: str
    _tabtooltip: str
    _pnlGeneralData: Any
    _pnlLogistics: Any
    _pnlMiscellaneous: Any
    _lst_callbacks: Any
    _lst_icons: Any
    _lst_mnu_labels: Any
    _lst_tooltips: Any
    def __init__(
        self, configuration: RAMSTKUserConfiguration, logger: RAMSTKLogManager
    ) -> None: ...
    def _do_request_make_comp_ref_des(self, __button: Gtk.ToolButton) -> None: ...
    _record_id: Any
    _parent_id: Any
    def _do_set_record_id(self, attributes: Dict[str, Any]) -> None: ...
    def __make_ui(self) -> None: ...

class HardwareAssessmentInputView(RAMSTKWorkView):
    _lst_title: List[str]
    _tag: str
    _tablabel: str
    _tabtooltip: str
    _dic_component_panels: Any
    _lst_callbacks: Any
    _lst_icons: Any
    _lst_tooltips: Any
    _pnlReliabilityInput: Any
    _pnlEnvironmentalInput: Any
    _pnlStressInput: Any
    _vpnLeft: Any
    _vpnRight: Any
    _hazard_rate_method_id: int
    _subcategory_id: int
    def __init__(
        self, configuration: RAMSTKUserConfiguration, logger: RAMSTKLogManager
    ) -> None: ...
    def _do_pack_component_panel(self, attributes: Dict[str, Any]) -> None: ...
    def _do_request_calculate(self, __button: Gtk.ToolButton) -> None: ...
    def _do_request_update(self, __button: Gtk.ToolButton) -> None: ...
    _record_id: Any
    _parent_id: Any
    def _do_set_record_id(self, attributes: Dict[str, Any]) -> None: ...
    def __make_ui(self) -> None: ...

class HardwareAssessmentResultsView(RAMSTKWorkView):
    _lst_title: Any
    _tag: str
    _tablabel: str
    _tabtooltip: str
    _dic_component_results: Any
    _lst_callbacks: Any
    _lst_icons: Any
    _lst_tooltips: Any
    _hazard_rate_method_id: int
    _subcategory_id: int
    _pnlAvailabilityResults: Any
    _pnlReliabilityResults: Any
    _pnlStressResults: Any
    _vpnLeft: Any
    _vpnRight: Any
    def __init__(
        self, configuration: RAMSTKUserConfiguration, logger: RAMSTKLogManager
    ) -> None: ...
    def _do_pack_component_panel(self, attributes: Dict[str, Any]) -> None: ...
    def _do_request_calculate(self, __button: Gtk.ToolButton) -> None: ...
    _record_id: Any
    _parent_id: Any
    def _do_set_record_id(self, attributes: Dict[str, Any]) -> None: ...
    def __make_ui(self) -> None: ...



# -*- coding: utf-8 -*-
#
#       ramstk.views.gtk3.usage_profile.view.py is part of The RAMSTK Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""GTK3 Usage Profile Views."""

# Standard Library Imports
from typing import Any, Dict

# Third Party Imports
from pubsub import pub

# RAMSTK Package Imports
from ramstk.configuration import RAMSTKUserConfiguration
from ramstk.logger import RAMSTKLogManager
from ramstk.views.gtk3 import Gtk, _
from ramstk.views.gtk3.widgets import RAMSTKListView

# RAMSTK Local Imports
from . import UsageProfileTreePanel


class UsageProfileListView(RAMSTKListView):
    """Display Usage Profiles associated with the selected Revision.

    The attributes of a Usage Profile List View are:

    :cvar _tag: the name of the module.

    :ivar _lst_mnu_labels: the list of labels for the view's pop-up
        menu.  The labels are listed in the order they appear in the menu.
    :ivar _lst_tooltips: the list of tooltips for the view's
        toolbar buttons and pop-up menu.  The tooltips are listed in the
        order they appear on the toolbar or pop-up menu.
    """

    # Define private dict class attributes.

    # Define private scalar class attributes.
    _tag: str = "usage_profile"
    _tablabel: str = "<span weight='bold'>" + _("Usage\nProfiles") + "</span>"
    _tabtooltip: str = _("Displays usage profiles for the selected revision.")

    # Define public dictionary class attributes.

    # Define public list class attributes.

    # Define public scalar class attributes.

    def __init__(
        self, configuration: RAMSTKUserConfiguration, logger: RAMSTKLogManager
    ) -> None:
        """Initialize an instance of the Usage Profile list view.

        :param configuration: the RAMSTK Configuration class instance.
        :param logger: the RAMSTKLogManager class instance.
        """
        super().__init__(configuration, logger)

        # Initialize private dictionary attributes.

        # Initialize private list attributes.
        self._lst_callbacks[0] = self._do_request_insert_sibling
        self._lst_callbacks.insert(1, self._do_request_insert_child)
        self._lst_callbacks.insert(2, self._do_request_delete)
        self._lst_icons[0] = "insert_sibling"
        self._lst_icons.insert(1, "insert_child")
        self._lst_icons.insert(2, "remove")
        self._lst_mnu_labels = [
            _("Add Sibling"),
            _("Add Child"),
            _("Delete Selected"),
            _("Save Selected"),
            _("Save Profile"),
        ]
        self._lst_tooltips = [
            _(
                "Add a new usage profile entity at the same level "
                "as the currently selected entity."
            ),
            _(
                "Add a new usage profile entity one level below the "
                "currently selected entity."
            ),
            _("Delete the currently selected entity from the usage profile."),
            _("Save changes to the currently selected entity in the usage profile."),
            _("Save changes to all entities in the usage profile."),
        ]

        # Initialize private scalar attributes.
        self._pnlPanel = UsageProfileTreePanel()

        # Initialize public dictionary attributes.

        # Initialize public list attributes.

        # Initialize public scalar attributes.

        self.__make_ui()

        # Subscribe to PyPubSub messages.
        pub.subscribe(self._do_set_record_id, "selected_mission")
        pub.subscribe(self._do_set_record_id, "selected_mission_phase")
        pub.subscribe(self._do_set_record_id, "selected_environment")

    # pylint: disable=unused-argument
    def _do_request_delete(self, __button: Gtk.ToolButton) -> None:
        """Request to delete the selected Usage Profile record.

        :param __button: the Gtk.ToolButton() that called this method.
        :return: None
        """
        _parent = self.get_parent().get_parent().get_parent().get_parent().get_parent()
        _dialog = super().do_raise_dialog(parent=_parent)
        _dialog.do_set_message(
            message=_(
                f"You are about to delete {self._tag} {self.dic_pkeys['record_id']} "
                f"and all data associated with it.  Is this really what you want to do?"
            )
        )
        _dialog.do_set_message_type(message_type="question")

        if _dialog.do_run() == Gtk.ResponseType.YES:
            super().do_set_cursor_busy()
            pub.sendMessage(
                f"request_delete_{self._tag}",
                node_id=self.dic_pkeys["record_id"],
            )

        _dialog.do_destroy()

    # pylint: disable=unused-argument
    def _do_request_insert_child(self, __button: Gtk.ToolButton) -> None:
        """Request to add an entity to the Usage Profile.

        :return: None
        """
        super().do_set_cursor_busy()

        _attributes = self.dic_pkeys

        if self._tag == "mission":
            _attributes.pop("environment_id")
            pub.sendMessage(
                "request_insert_mission_phase",
                attributes=_attributes,
            )
        elif self._tag == "mission_phase":
            pub.sendMessage(
                "request_insert_environment",
                attributes=_attributes,
            )
        elif self._tag == "environment":
            _error = _("An environmental condition cannot have a child.")
            _parent = (
                self.get_parent().get_parent().get_parent().get_parent().get_parent()
            )
            _dialog = super().do_raise_dialog(parent=_parent)
            _dialog.do_set_message(message=_error)
            _dialog.do_set_message_type(message_type="error")
            _dialog.do_run()
            _dialog.do_destroy()
            pub.sendMessage(
                "fail_insert_usage_profile",
                error_message=_error,
            )

    # pylint: disable=unused-argument
    def _do_request_insert_sibling(self, __button: Gtk.ToolButton) -> None:
        """Request to add a sibling entity to the Usage Profile.

        :return: None
        """
        super().do_set_cursor_busy()

        _attributes = self.dic_pkeys
        if self._tag == "mission":
            _attributes.pop("phase_id")
            _attributes.pop("environment_id")
        elif self._tag == "mission_phase":
            _attributes.pop("environment_id")

        pub.sendMessage(
            f"request_insert_{self._tag}",
            attributes=_attributes,
        )

    def _do_set_record_id(self, attributes: Dict[str, Any]) -> None:
        """Set the usage profile's record ID.

        :param attributes: the attributes dict for the selected usage profile element.
        :return: None
        :rtype: None
        """
        self.dic_pkeys["revision_id"] = attributes["revision_id"]
        self.dic_pkeys["phase_id"] = attributes["phase_id"]
        self.dic_pkeys["environment_id"] = attributes["environment_id"]
        self.dic_pkeys["mission_id"] = attributes["mission_id"]

        if (
            attributes["mission_id"]
            * attributes["phase_id"]
            * attributes["environment_id"]
        ) > 0:
            self.dic_pkeys["parent_id"] = attributes["phase_id"]
            self.dic_pkeys["record_id"] = attributes["environment_id"]
            self._tag = "environment"
        elif (attributes["mission_id"] * attributes["phase_id"]) > 0:
            self.dic_pkeys["parent_id"] = attributes["mission_id"]
            self.dic_pkeys["record_id"] = attributes["phase_id"]
            self._tag = "mission_phase"
        else:
            self.dic_pkeys["parent_id"] = 0
            self.dic_pkeys["record_id"] = attributes["mission_id"]
            self._tag = "mission"

    def __make_ui(self):
        """Build the user interface for the usage profile list view.

        :return: None
        """
        super().make_ui()

        self._pnlPanel.dic_units = (
            self.RAMSTK_USER_CONFIGURATION.RAMSTK_MEASUREMENT_UNITS
        )
        self._pnlPanel.do_load_comboboxes()

        self._pnlPanel.tvwTreeView.dic_handler_id[
            "button-press"
        ] = self._pnlPanel.tvwTreeView.connect(
            "button_press_event", super().on_button_press
        )
        for _element in ["mission", "mission_phase", "environment"]:
            self._pnlPanel.dic_icons[_element] = self._dic_icons[_element]



# -*- coding: utf-8 -*-
#
#       ramstk.views.gtk3.stakeholder.view.py is part of the RAMSTK Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""GTK3 Stakeholder Views."""

# Standard Library Imports
from typing import Any, Dict

# Third Party Imports
from pubsub import pub

# RAMSTK Package Imports
from ramstk.configuration import RAMSTKUserConfiguration
from ramstk.logger import RAMSTKLogManager
from ramstk.views.gtk3 import Gtk, _
from ramstk.views.gtk3.widgets import RAMSTKListView

# RAMSTK Local Imports
from . import StakeholderTreePanel


class StakeholderListView(RAMSTKListView):
    """Display Stakeholder Inputs associated with the selected Revision.

    The Stakeholder List View displays all the stakeholder inputs associated
    with the selected Requirement.  The attributes of the Stakeholder List
    View are:

    :cvar _tag: the name of the module.

    :ivar _lst_callbacks: the list of callback methods for the view's
        toolbar buttons and pop-up menu.  The methods are listed in the order
        they appear on the toolbar and pop-up menu.
    :ivar _lst_icons: the list of icons for the view's toolbar buttons
        and pop-up menu.  The icons are listed in the order they appear on the
        toolbar and pop-up menu.
    :ivar _lst_mnu_labels: the list of labels for the view's pop-up
        menu.  The labels are listed in the order they appear in the menu.
    :ivar _lst_tooltips: the list of tooltips for the view's
        toolbar buttons and pop-up menu.  The tooltips are listed in the
        order they appear on the toolbar or pop-up menu.
    """

    # Define private dictionary class attributes.

    # Define private list class attributes.

    # Define private scalar class attributes.
    _tag = "stakeholder"
    _tablabel = "<span weight='bold'>" + _("Stakeholder\nInputs") + "</span>"
    _tabtooltip = _("Displays stakeholder inputs for the selected revision.")
    _view_type = "list"

    # Define public dictionary class attributes.

    # Define public list class attributes.

    # Define public scalar class attributes.

    def __init__(
        self, configuration: RAMSTKUserConfiguration, logger: RAMSTKLogManager
    ) -> None:
        """Initialize the List View for the Requirement package.

        :param configuration: the RAMSTK Configuration class instance.
        :param logger: the RAMSTKLogManager class instance.
        """
        super().__init__(configuration, logger)

        # Initialize private dictionary attributes.

        # Initialize private list attributes.
        self._lst_callbacks.insert(1, self._do_request_delete)
        self._lst_callbacks.insert(2, self._do_request_calculate)
        self._lst_icons.insert(1, "remove")
        self._lst_icons.insert(2, "calculate")
        self._lst_mnu_labels.insert(1, _("Delete Selected Input"))
        self._lst_mnu_labels.insert(2, _("Calculate Inputs"))
        self._lst_tooltips.insert(0, _("Add a new stakeholder input."))
        self._lst_tooltips.insert(
            1, _("Remove the currently selected stakeholder input.")
        )
        self._lst_tooltips.insert(
            2, _("Calculate the stakeholder improvement factors.")
        )
        self._lst_tooltips.insert(
            3, _("Update the currently selected stakeholder input.")
        )
        self._lst_tooltips.insert(4, _("Update all stakeholder inputs."))

        # Initialize private scalar attributes.
        self._pnlPanel = StakeholderTreePanel()

        # Initialize public dictionary attributes.

        # Initialize public list attributes.

        # Initialize public scalar attributes.

        self.__make_ui()

        # Subscribe to PyPubSub messages.
        pub.subscribe(self._do_set_record_id, "selected_stakeholder")

    def _do_add_to_affinity_group(self, new_text: str) -> None:
        """Add an entry to the RAMSTK_AFFINITY_GROUP dictionary.

        :param new_text: the name of the new group to add to the
            RAMSTK_AFFINITY_GROUP dictionary.
        :return: None
        """
        try:
            _new_key = (
                max(self.RAMSTK_USER_CONFIGURATION.RAMSTK_AFFINITY_GROUPS.keys()) + 1
            )
        except ValueError:
            _new_key = 1
        self.RAMSTK_USER_CONFIGURATION.RAMSTK_AFFINITY_GROUPS[_new_key] = str(new_text)

    # pylint: disable=unused-argument
    def _do_request_calculate(self, __button: Gtk.ToolButton) -> None:
        """Request to calculate the selected Stakeholder input.

        :param __button: the Gtk.ToolButton() that called this method.
        :return: None
        """
        super().do_set_cursor_busy()
        pub.sendMessage(
            "request_calculate_stakeholder", node_id=self.dic_pkeys["record_id"]
        )
        super().do_set_cursor_active()

    # pylint: disable=unused-argument
    def _do_request_calculate_all(self, __button: Gtk.ToolButton) -> None:
        """Request to calculate all Stakeholder inputs.

        :param __button: the Gtk.ToolButton() that called this method.
        :return: None
        """
        super().do_set_cursor_busy()
        pub.sendMessage("request_calculate_all_stakeholders")
        super().do_set_cursor_active()

    # pylint: disable=unused-argument
    def _do_request_delete(self, __button: Gtk.ToolButton) -> None:
        """Request to delete the selected Stakeholder.

        :param __button: the Gtk.ToolButton() that called this method.
        :return: None
        """
        _parent = self.get_parent().get_parent().get_parent().get_parent().get_parent()
        _dialog = super().do_raise_dialog(parent=_parent)
        _dialog.do_set_message(
            message=_(
                "You are about to delete Stakeholder input {0:d} and "
                "all data associated with it.  Is this really what you "
                "want to do?"
            ).format(self.dic_pkeys["record_id"])
        )
        _dialog.do_set_message_type(message_type="question")

        if _dialog.do_run() == Gtk.ResponseType.YES:
            super().do_set_cursor_busy()
            pub.sendMessage(
                "request_delete_stakeholder", node_id=self.dic_pkeys["record_id"]
            )

        _dialog.do_destroy()

    def _do_set_record_id(self, attributes: Dict[str, Any]) -> None:
        """Set the stakeholder input's record ID.

        :param attributes: the attributes dict for the selected stakeholder
            input.
        :return: None
        :rtype: None
        """
        self.dic_pkeys["revision_id"] = attributes["revision_id"]
        self.dic_pkeys["parent_id"] = 0
        self.dic_pkeys["stakeholder_id"] = attributes["requirement_id"]
        self.dic_pkeys["record_id"] = attributes["stakeholder_id"]

    def __make_ui(self) -> None:
        """Build the user interface for the stakeholder input list view.

        :return: None
        """
        super().make_ui()

        self._pnlPanel.do_load_affinity_groups(
            self.RAMSTK_USER_CONFIGURATION.RAMSTK_AFFINITY_GROUPS
        )
        self._pnlPanel.do_load_stakeholders(
            self.RAMSTK_USER_CONFIGURATION.RAMSTK_STAKEHOLDERS
        )

        self._pnlPanel.tvwTreeView.dic_handler_id[
            "button-press"
        ] = self._pnlPanel.tvwTreeView.connect(
            "button_press_event", super().on_button_press
        )



# pylint: disable=cyclic-import
# -*- coding: utf-8 -*-
#
#       ramstk.models.similar_item.table.py is part of The RAMSTK Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""Similar Item Table Model."""

# Standard Library Imports
from collections import OrderedDict
from typing import Any, Dict, Type

# Third Party Imports
from pubsub import pub

# RAMSTK Package Imports
from ramstk.analyses import similaritem
from ramstk.models import RAMSTKBaseTable, RAMSTKSimilarItemRecord


class RAMSTKSimilarItemTable(RAMSTKBaseTable):
    """Contain the attributes and methods of the Similar Item data manager."""

    # Define private dictionary class attributes.

    # Define private list class attributes.

    # Define private scalar class attributes.
    _db_id_colname = "fld_hardware_id"
    _db_tablename = "ramstk_similar_item"
    _select_msg = "selected_revision"
    _tag = "similar_item"

    # Define public dictionary class attributes.

    # Define public list class attributes.

    # Define public scalar class attributes.

    def __init__(self, **kwargs: Dict[str, Any]) -> None:
        """Initialize a Hardware data manager instance."""
        super().__init__(**kwargs)

        # Initialize private dictionary attributes.

        # Initialize private list attributes.
        self._lst_id_columns = [
            "revision_id",
            "hardware_id",
        ]

        # Initialize private scalar attributes.
        self._node_hazard_rate: float = 0.0
        self._record: Type[RAMSTKSimilarItemRecord] = RAMSTKSimilarItemRecord

        # Initialize public dictionary attributes.

        # Initialize public list attributes.

        # Initialize public scalar attributes.
        self.pkey = "hardware_id"

        # Subscribe to PyPubSub messages.
        pub.subscribe(self.do_calculate_similar_item, "request_calculate_similar_item")
        pub.subscribe(
            self.do_roll_up_change_descriptions, "request_roll_up_change_descriptions"
        )

    def do_get_new_record(  # pylint: disable=method-hidden
        self, attributes: Dict[str, Any]
    ) -> object:
        """Gets a new record instance with attributes set.

        :param attributes: the dict of attribute values to assign to the new record.
        :return: None
        :rtype: None
        """
        self._parent_id = attributes["parent_id"]

        _new_record = self._record()
        _new_record.revision_id = attributes["revision_id"]
        _new_record.hardware_id = attributes["hardware_id"]
        _new_record.parent_id = attributes["parent_id"]

        return _new_record

    def do_calculate_similar_item(self, node_id: int) -> None:
        """Perform a similar item calculation for record ID.

        :param node_id: the node (similar item) ID to calculate.
        :return: None
        :rtype: None
        """
        _dic_method = {
            1: self._do_calculate_topic_633,
            2: self._do_calculate_user_defined,
        }

        _record = self.tree.get_node(node_id).data[self._tag]
        try:
            _method = _dic_method[_record.similar_item_method_id]
            _method(node_id)

            pub.sendMessage(
                "succeed_calculate_similar_item",
                tree=self.tree,
            )
        except KeyError:
            _error_msg: str = (
                "Failed to calculate similar item reliability for hardware ID {0}.  "
                "Unknown similar item method ID {1} selected.".format(
                    node_id, _record.similar_item_method_id
                )
            )
            pub.sendMessage(
                "do_log_debug",
                logger_name="DEBUG",
                message=_error_msg,
            )
            pub.sendMessage(
                "fail_calculate_similar_item",
                error_message=_error_msg,
            )

    def do_roll_up_change_descriptions(self, node_id: int) -> None:
        """Concatenate child change descriptions for the node ID similar item.

        :param node_id: the record ID of the parent to which the rolled-up
            descriptions are assigned.
        :return: None
        :rtype: None
        """
        _change_description_1 = ""
        _change_description_2 = ""
        _change_description_3 = ""
        _change_description_4 = ""
        _change_description_5 = ""
        _change_description_6 = ""
        _change_description_7 = ""
        _change_description_8 = ""
        _change_description_9 = ""
        _change_description_10 = ""

        for _node in self.tree.children(node_id):
            _change_description_1 += (
                _node.data["similar_item"].change_description_1 + "\n\n"
            )
            _change_description_2 += (
                _node.data["similar_item"].change_description_2 + "\n\n"
            )
            _change_description_3 += (
                _node.data["similar_item"].change_description_3 + "\n\n"
            )
            _change_description_4 += (
                _node.data["similar_item"].change_description_4 + "\n\n"
            )
            _change_description_5 += (
                _node.data["similar_item"].change_description_5 + "\n\n"
            )
            _change_description_6 += (
                _node.data["similar_item"].change_description_6 + "\n\n"
            )
            _change_description_7 += (
                _node.data["similar_item"].change_description_7 + "\n\n"
            )
            _change_description_8 += (
                _node.data["similar_item"].change_description_8 + "\n\n"
            )
            _change_description_9 += (
                _node.data["similar_item"].change_description_9 + "\n\n"
            )
            _change_description_10 += (
                _node.data["similar_item"].change_description_10 + "\n\n"
            )

        self.do_set_attributes(
            node_id=[node_id],
            package={
                "change_description_1": _change_description_1,
            },
        )
        self.do_set_attributes(
            node_id=[node_id],
            package={
                "change_description_2": _change_description_2,
            },
        )
        self.do_set_attributes(
            node_id=[node_id],
            package={
                "change_description_3": _change_description_3,
            },
        )
        self.do_set_attributes(
            node_id=[node_id],
            package={
                "change_description_4": _change_description_4,
            },
        )
        self.do_set_attributes(
            node_id=[node_id],
            package={
                "change_description_5": _change_description_5,
            },
        )
        self.do_set_attributes(
            node_id=[node_id],
            package={
                "change_description_6": _change_description_6,
            },
        )
        self.do_set_attributes(
            node_id=[node_id],
            package={
                "change_description_7": _change_description_7,
            },
        )
        self.do_set_attributes(
            node_id=[node_id],
            package={
                "change_description_8": _change_description_8,
            },
        )
        self.do_set_attributes(
            node_id=[node_id],
            package={
                "change_description_9": _change_description_9,
            },
        )
        self.do_set_attributes(
            node_id=[node_id],
            package={
                "change_description_10": _change_description_10,
            },
        )

        pub.sendMessage(
            "succeed_roll_up_change_descriptions",
            tree=self.tree,
        )

    def _do_calculate_topic_633(self, node_id: int) -> None:
        """Calculate the similar item hazard rate per topic 6.3.3.

        .. note:: this analysis uses the adjustment factors from RAC/RiAC's The
            Reliability Toolkit, Commercial Practices Edition, section 6.3.3.

        :param node_id: the record ID to calculate similar item reliability.
        :return: None
        :rtype: None
        """
        _attributes = self.tree.get_node(node_id).data[self._tag].get_attributes()

        _environment = {
            "from": _attributes["environment_from_id"],
            "to": _attributes["environment_to_id"],
        }
        _quality = {
            "from": _attributes["quality_from_id"],
            "to": _attributes["quality_to_id"],
        }
        _temperature = {
            "from": _attributes["temperature_from"],
            "to": _attributes["temperature_to"],
        }

        (
            _attributes["change_factor_1"],
            _attributes["change_factor_2"],
            _attributes["change_factor_3"],
            _attributes["result_1"],
        ) = similaritem.calculate_topic_633(
            _environment, _quality, _temperature, self._node_hazard_rate
        )

        self.do_set_attributes_all(
            attributes=_attributes,
        )

    def _do_calculate_user_defined(self, node_id: int) -> None:
        """Calculate the user-defined similar item hazard rate.

        :param node_id: the record ID to calculate similar item reliability.
        :return: None
        :rtype: None
        """
        _attributes = self.tree.get_node(node_id).data[self._tag].get_attributes()

        _sia: Dict[str, Any] = OrderedDict(
            {
                _key: None
                for _key in [
                    "hr",
                    "pi1",
                    "pi2",
                    "pi3",
                    "pi3",
                    "pi4",
                    "pi5",
                    "pi6",
                    "pi7",
                    "pi8",
                    "pi9",
                    "pi10",
                    "uf1",
                    "uf2",
                    "uf3",
                    "uf4",
                    "uf5",
                    "ui1",
                    "ui2",
                    "ui3",
                    "ui4",
                    "ui5",
                    "equation1",
                    "equation2",
                    "equation3",
                    "equation4",
                    "equation5",
                    "res1",
                    "res2",
                    "res3",
                    "res4",
                    "res5",
                ]
            }
        )

        _sia["hr"] = self._node_hazard_rate

        _sia = similaritem.set_user_defined_change_factors(
            _sia,
            [
                _attributes["change_factor_1"],
                _attributes["change_factor_2"],
                _attributes["change_factor_3"],
                _attributes["change_factor_4"],
                _attributes["change_factor_5"],
                _attributes["change_factor_6"],
                _attributes["change_factor_7"],
                _attributes["change_factor_8"],
                _attributes["change_factor_9"],
                _attributes["change_factor_10"],
            ],
        )

        _sia = similaritem.set_user_defined_floats(
            _sia,
            [
                _attributes["user_float_1"],
                _attributes["user_float_2"],
                _attributes["user_float_3"],
                _attributes["user_float_4"],
                _attributes["user_float_5"],
            ],
        )

        _sia = similaritem.set_user_defined_ints(
            _sia,
            [
                _attributes["user_int_1"],
                _attributes["user_int_2"],
                _attributes["user_int_3"],
                _attributes["user_int_4"],
                _attributes["user_int_5"],
            ],
        )

        _sia = similaritem.set_user_defined_functions(
            _sia,
            [
                _attributes["function_1"],
                _attributes["function_2"],
                _attributes["function_3"],
                _attributes["function_4"],
                _attributes["function_5"],
            ],
        )

        _sia = similaritem.set_user_defined_results(
            _sia,
            [
                _attributes["result_1"],
                _attributes["result_2"],
                _attributes["result_3"],
                _attributes["result_4"],
                _attributes["result_5"],
            ],
        )

        _sia = similaritem.calculate_user_defined(_sia)
        _attributes["result_1"] = _sia["res1"]
        _attributes["result_2"] = _sia["res2"]
        _attributes["result_3"] = _sia["res3"]
        _attributes["result_4"] = _sia["res4"]
        _attributes["result_5"] = _sia["res5"]

        self.do_set_attributes_all(
            attributes=_attributes,
        )



# -*- coding: utf-8 -*-
#
#       ramstk.views.gtk3.failure_definition.view.py is part of The RAMSTK Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""GTK3 Failure Definition Views."""

# Standard Library Imports
from typing import Any, Dict

# Third Party Imports
from pubsub import pub

# RAMSTK Package Imports
from ramstk.configuration import RAMSTKUserConfiguration
from ramstk.logger import RAMSTKLogManager
from ramstk.views.gtk3 import Gtk, _
from ramstk.views.gtk3.widgets import RAMSTKListView

# RAMSTK Local Imports
from . import FailureDefinitionTreePanel


class FailureDefinitionListView(RAMSTKListView):
    """Display failure definitions associated with the selected revision.

    The attributes of the failure definition list view are:

    :cvar _tag: the name of the module.  Spaces are replaced with single
        underscores if the name is more than one word.
    :cvar _tablabel: the text to display on the view's tab.
    :cvar _tabtooltip: the text to display in the tooltip for the view's tab.

    :ivar _lst_mnu_labels: the list of labels for the view's pop-up
        menu.  The labels are listed in the order they appear in the menu.
    :ivar _lst_tooltips: the list of tooltips for the view's
        toolbar buttons and pop-up menu.  The tooltips are listed in the
        order they appear on the toolbar or pop-up menu.
    """

    # Define private dictionary class attributes.

    # Define private list class attributes.

    # Define private scalar class attributes.
    _tag: str = "failure_definition"
    _tablabel = "<span weight='bold'>" + _("Failure\nDefinitions") + "</span>"
    _tabtooltip = _("Displays failure definitions for the selected revision.")

    # Define public dictionary class attributes.

    # Define public list class attributes.

    # Define public scalar class attributes.

    def __init__(
        self, configuration: RAMSTKUserConfiguration, logger: RAMSTKLogManager
    ) -> None:
        """Initialize the List View for the failure definition package.

        :param configuration: the RAMSTK Configuration class instance.
        :param logger: the RAMSTKLogManager class instance.
        """
        super().__init__(configuration, logger)

        # Initialize private dictionary attributes.

        # Initialize private list attributes.
        self._lst_callbacks.insert(1, self._do_request_delete)
        self._lst_icons.insert(1, "remove")
        self._lst_mnu_labels = [
            _("Add Definition"),
            _("Delete Selected Definition"),
            _("Save Selected Definition"),
            _("Save All Definitions"),
        ]
        self._lst_tooltips = [
            _("Add a new failure definition."),
            _("Delete the currently selected failure definition."),
            _("Save changes to the currently selected failure definition"),
            _("Save changes to all failure definitions."),
        ]

        # Initialize private scalar attributes.
        self._pnlPanel = FailureDefinitionTreePanel()

        # Initialize public dictionary attributes.

        # Initialize public list attributes.

        # Initialize public scalar attributes.

        self.__make_ui()

        # Subscribe to PyPubSub messages.
        pub.subscribe(self._do_set_record_id, "selected_failure_definition")

    # pylint: disable=unused-argument
    def _do_request_delete(self, __button: Gtk.ToolButton) -> None:
        """Request to delete the selected Failure Definition record.

        :param __button: the Gtk.ToolButton() that called this method.
        :return: None
        """
        _parent = self.get_parent().get_parent().get_parent().get_parent().get_parent()
        _dialog = super().do_raise_dialog(parent=_parent)
        _dialog.do_set_message(
            message=_(
                f"You are about to delete Failure Definition "
                f"{self.dic_pkeys['record_id']} and all data associated with it.  Is "
                f"this really what you want to do?"
            )
        )
        _dialog.do_set_message_type(message_type="question")

        if _dialog.do_run() == Gtk.ResponseType.YES:
            super().do_set_cursor_busy()
            pub.sendMessage(
                "request_delete_failure_definition",
                node_id=self.dic_pkeys["record_id"],
            )

        _dialog.do_destroy()

    def _do_set_record_id(self, attributes: Dict[str, Any]) -> None:
        """Set the failure definition's record ID.

        :param attributes: the attributes dict for the selected failure
            definition.
        :return: None
        :rtype: None
        """
        self.dic_pkeys["revision_id"] = attributes["revision_id"]
        self.dic_pkeys["definition_id"] = attributes["definition_id"]
        self.dic_pkeys["parent_id"] = 0
        self.dic_pkeys["record_id"] = attributes["definition_id"]

    def __make_ui(self) -> None:
        """Build the user interface for the failure definition list view.

        :return: None
        """
        super().make_ui()

        self._pnlPanel.tvwTreeView.dic_handler_id[
            "button-press"
        ] = self._pnlPanel.tvwTreeView.connect(
            "button_press_event", super().on_button_press
        )



# Standard Library Imports
from typing import Any, Dict

# RAMSTK Package Imports
from ramstk.models import RAMSTKActionRecord as RAMSTKActionRecord
from ramstk.models import RAMSTKBaseTable as RAMSTKBaseTable

class RAMSTKActionTable(RAMSTKBaseTable):
    _db_id_colname: str = ...
    _db_tablename: str = ...
    _select_msg: str = ...
    _tag: str = ...
    _lst_id_columns: Any = ...
    _record: Any = ...
    pkey: str = ...
    def __init__(self, **kwargs: Dict[str, Any]) -> None: ...
    def do_get_new_record(self, attributes: Dict[str, Any]) -> object: ...



# Standard Library Imports
from typing import Any, Dict

# RAMSTK Package Imports
from ramstk.views.gtk3 import GObject as GObject
from ramstk.views.gtk3 import Gtk as Gtk

# RAMSTK Local Imports
from .label import RAMSTKLabel as RAMSTKLabel

class RAMSTKFrame(Gtk.Frame):
    def __init__(self) -> None: ...
    def do_set_properties(self, **kwargs: Any) -> None: ...



# -*- coding: utf-8 -*-
#
#       ramstk.logger.py is part of The RAMSTK Project
#
# All rights reserved.
# Copyright 2019 Doyle Rowland doyle.rowland <AT> reliaqual <DOT> com
"""RAMSTK Logger Module."""

# Standard Library Imports
import logging
import sys
from typing import Dict

# Third Party Imports
from pubsub import pub

LOGFORMAT = logging.Formatter("%(asctime)s - %(name)s - %(lineno)s : %(message)s")


class RAMSTKLogManager:
    """Class to manage logging of RAMSTK messages."""

    loggers: Dict[str, logging.Logger] = {}

    def __init__(self, log_file: str) -> None:
        """Initialize an instance of the LogManager.

        :param log_file: the absolute path to the log file to use with this
            log manager.
        """
        # Initialize private dictionary attributes.

        # Initialize private list attributes.

        # Initialize private scalar attributes.

        # Initialize public dictionary attributes.

        # Initialize public list attributes.

        # Initialize public scalar attributes.
        self.log_file = log_file

        # Subscribe to PyPubSub messages.
        pub.subscribe(self._do_log_fail_message, "fail_connect_program_database")
        pub.subscribe(self._do_log_fail_message, "fail_delete_environment")
        pub.subscribe(self._do_log_fail_message, "fail_delete_failure_definition")
        pub.subscribe(self._do_log_fail_message, "fail_delete_fmea")
        pub.subscribe(self._do_log_fail_message, "fail_delete_function")
        pub.subscribe(self._do_log_fail_message, "fail_delete_hazard")
        pub.subscribe(self._do_log_fail_message, "fail_delete_mission")
        pub.subscribe(self._do_log_fail_message, "fail_delete_mission_phase")
        pub.subscribe(self._do_log_fail_message, "fail_delete_revision")
        pub.subscribe(self._do_log_fail_message, "fail_import_module")
        pub.subscribe(self._do_log_fail_message, "fail_insert_action")
        pub.subscribe(self._do_log_fail_message, "fail_insert_cause")
        pub.subscribe(self._do_log_fail_message, "fail_insert_control")
        pub.subscribe(self._do_log_fail_message, "fail_insert_environment")
        pub.subscribe(self._do_log_fail_message, "fail_insert_failure_definition")
        pub.subscribe(self._do_log_fail_message, "fail_insert_mechanism")
        pub.subscribe(self._do_log_fail_message, "fail_insert_mission")
        pub.subscribe(self._do_log_fail_message, "fail_insert_mission_phase")
        pub.subscribe(self._do_log_fail_message, "fail_insert_mode")
        pub.subscribe(self._do_log_fail_message, "fail_insert_function")
        pub.subscribe(self._do_log_fail_message, "fail_insert_hazard")
        pub.subscribe(self._do_log_fail_message, "fail_insert_hardware")
        pub.subscribe(self._do_log_fail_message, "fail_insert_validation")
        pub.subscribe(self._do_log_fail_message, "fail_insert_stakeholder")
        pub.subscribe(self._do_log_fail_message, "fail_insert_revision")
        pub.subscribe(self._do_log_fail_message, "fail_insert_requirement")
        pub.subscribe(self._do_log_fail_message, "fail_insert_opload")
        pub.subscribe(self._do_log_fail_message, "fail_insert_opstress")
        pub.subscribe(self._do_log_fail_message, "fail_insert_record")
        pub.subscribe(self._do_log_fail_message, "fail_insert_test_method")
        pub.subscribe(self._do_log_fail_message, "fail_update_fmea")
        pub.subscribe(self._do_log_fail_message, "fail_update_function")
        pub.subscribe(self._do_log_fail_message, "fail_update_hardware")
        pub.subscribe(self._do_log_fail_message, "fail_update_record")
        pub.subscribe(self._do_log_fail_message, "fail_update_requirement")
        pub.subscribe(self._do_log_fail_message, "fail_update_revision")

        pub.subscribe(self.do_log_debug, "do_log_debug_msg")
        pub.subscribe(self.do_log_info, "do_log_info_msg")
        pub.subscribe(self.do_log_warning, "do_log_warning_msg")
        pub.subscribe(self.do_log_error, "do_log_error_msg")
        pub.subscribe(self.do_log_critical, "do_log_critical_msg")

        # Create a logger for the pypubsub fail_* messages.
        self.do_create_logger(__name__, "WARN")

    def _do_log_fail_message(self, error_message: str) -> None:
        """Log PyPubSub broadcast fail messages.

        :param error_message: the error message that was part of the
            broadcast package.
        :return: None
        :rtype: None
        """
        self.loggers[__name__].warning(error_message)

    @staticmethod
    def _get_console_handler(log_level: str) -> logging.Handler:
        """Create the log handler for console output.

        :return: _c_handler
        :rtype: :class:`logging.Handler`
        """
        _c_handler = logging.StreamHandler(sys.stdout)
        _c_handler.setLevel(log_level)
        _c_handler.setFormatter(LOGFORMAT)

        return _c_handler

    def _get_file_handler(self, log_level: str) -> logging.Handler:
        """Create the log handler for file output.

        :return: _f_handler
        :rtype: :class:`logging.Handler`
        """
        _f_handler = logging.FileHandler(self.log_file)
        _f_handler.setLevel(log_level)
        _f_handler.setFormatter(LOGFORMAT)

        return _f_handler

    def do_create_logger(
        self, logger_name: str, log_level: str, to_tty: bool = False
    ) -> None:
        """Create a logger instance.

        :param logger_name: the name of the logger used in the application.
        :param log_level: the level of messages to log.
        :param to_tty: boolean indicating whether this logger will
            also dump messages to the terminal.
        :return: None
        :rtype: None
        """
        _logger = logging.getLogger(logger_name)
        _logger.setLevel(log_level)

        _logger.addHandler(self._get_file_handler(log_level))
        if to_tty:
            _logger.addHandler(self._get_console_handler(log_level))

        self.loggers[logger_name] = _logger

    def do_log_debug(self, logger_name: str, message: str) -> None:
        """Log DEBUG level messages.

        :param logger_name: the name of the logger used in the application.
        :param message: the message to log.
        :return: None
        :rtype: None
        """
        if self.loggers[logger_name].isEnabledFor(logging.DEBUG):
            self.loggers[logger_name].debug(message)

    def do_log_exception(self, logger_name: str, exception: object) -> None:
        """Log EXCEPTIONS.

        :param logger_name: the name of the logger used in the application.
        :param exception: the exception to log.
        :return: None
        :rtype: None
        """
        if self.loggers[logger_name].isEnabledFor(logging.WARNING):
            self.loggers[logger_name].exception(exception)

    def do_log_info(self, logger_name: str, message: str) -> None:
        """Log INFO level messages.

        :param logger_name: the name of the logger used in the application.
        :param message: the message to log.
        :return: None
        :rtype: None
        """
        if self.loggers[logger_name].isEnabledFor(logging.INFO):
            self.loggers[logger_name].info(message)

    def do_log_warning(self, logger_name: str, message: str) -> None:
        """Log WARN level messages.

        :param logger_name: the name of the logger used in the application.
        :param message: the message to log.
        :return: None
        :rtype: None
        """
        if self.loggers[logger_name].isEnabledFor(logging.WARNING):
            self.loggers[logger_name].warning(message)

    def do_log_error(self, logger_name: str, message: str) -> None:
        """Log ERROR level messages.

        :param logger_name: the name of the logger used in the application.
        :param message: the message to log.
        :return: None
        :rtype: None
        """
        if self.loggers[logger_name].isEnabledFor(logging.ERROR):
            self.loggers[logger_name].error(message)

    def do_log_critical(self, logger_name: str, message: str) -> None:
        """Log CRITICAL level messages.

        :param logger_name: the name of the logger used in the application.
        :param message: the message to log.
        :return: None
        :rtype: None
        """
        self.loggers[logger_name].critical(message)



# -*- coding: utf-8 -*-
#
#       ramstk.models.commondb.RAMSTKHazards.py is part of The RAMSTK Project
#
# All rights reserved.
# Copyright 2007 - 2017 Doyle Rowland doyle.rowland <AT> reliaqual <DOT> com
"""RAMSTKHazard Table Module."""

# Third Party Imports
from sqlalchemy import Column, Integer, String

# RAMSTK Package Imports
from ramstk.db import RAMSTK_BASE
from ramstk.models import RAMSTKBaseRecord


class RAMSTKHazards(RAMSTK_BASE, RAMSTKBaseRecord):
    """Class to represent ramstk_hazard in the RAMSTK Common database."""

    __defaults__ = {
        "hazard_category": "Hazard Category",
        "hazard_subcategory": "Hazard Subcategory",
    }
    __tablename__ = "ramstk_hazards"
    __table_args__ = {"extend_existing": True}

    hazard_id = Column(
        "fld_hazard_id",
        Integer,
        primary_key=True,
        autoincrement=True,
        nullable=False,
    )
    hazard_category = Column(
        "fld_hazard_category", String(512), default="Hazard Category"
    )
    hazard_subcategory = Column(
        "fld_hazard_subcategory", String(512), default="Hazard Subcategory"
    )

    def get_attributes(self):
        """Retrieve current values of the RAMSTKHazard data model attributes.

        :return: {hazard_id, category, subcategory} pairs
        :rtype: tuple
        """
        _attributes = {
            "hazard_id": self.hazard_id,
            "hazard_category": self.hazard_category,
            "hazard_subcategory": self.hazard_subcategory,
        }

        return _attributes



# pylint: disable=duplicate-code
# -*- coding: utf-8 -*-
#
#       ramstk.models.mode.record.py is part of The RAMSTK Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy Rowland doyle.rowland <AT> reliaqual <DOT> com
"""Mode Record Model."""

# Third Party Imports
from sqlalchemy import Column, Float, ForeignKeyConstraint, Integer, String
from sqlalchemy.orm import relationship

# RAMSTK Package Imports
from ramstk.db import RAMSTK_BASE
from ramstk.models import RAMSTKBaseRecord


class RAMSTKModeRecord(RAMSTK_BASE, RAMSTKBaseRecord):
    """Class to represent table ramstk_mode in the RAMSTK Program database.

    This table shares a Many-to-One relationship with ramstk_function. This table
    shares a Many-to-One relationship with ramstk_hardware. This table shares a
    One-to-Many relationship with ramstk_mechanism.
    """

    __defaults__ = {
        "critical_item": 0,
        "description": "",
        "design_provisions": "",
        "detection_method": "",
        "effect_end": "",
        "effect_local": "",
        "effect_next": "",
        "effect_probability": 0.0,
        "hazard_rate_source": "",
        "isolation_method": "",
        "mission": "Default Mission",
        "mission_phase": "",
        "mode_criticality": 0.0,
        "mode_hazard_rate": 0.0,
        "mode_op_time": 0.0,
        "mode_probability": "",
        "mode_ratio": 0.0,
        "operator_actions": "",
        "other_indications": "",
        "remarks": "",
        "rpn_severity": 1,
        "rpn_severity_new": 1,
        "severity_class": "",
        "single_point": 0,
        "type_id": 0,
    }
    __tablename__ = "ramstk_mode"
    __table_args__ = (
        ForeignKeyConstraint(
            ["fld_revision_id", "fld_hardware_id"],
            ["ramstk_hardware.fld_revision_id", "ramstk_hardware.fld_hardware_id"],
        ),
        {"extend_existing": True},
    )

    revision_id = Column("fld_revision_id", Integer, primary_key=True, nullable=False)
    hardware_id = Column(
        "fld_hardware_id", Integer, primary_key=True, default=-1, nullable=False
    )
    mode_id = Column(
        "fld_mode_id", Integer, primary_key=True, autoincrement=True, nullable=False
    )

    critical_item = Column(
        "fld_critical_item", Integer, default=__defaults__["critical_item"]
    )
    description = Column(
        "fld_description", String(512), default=__defaults__["description"]
    )
    design_provisions = Column(
        "fld_design_provisions", String, default=__defaults__["design_provisions"]
    )
    detection_method = Column(
        "fld_detection_method", String(512), default=__defaults__["detection_method"]
    )
    effect_end = Column(
        "fld_effect_end", String(512), default=__defaults__["effect_end"]
    )
    effect_local = Column(
        "fld_effect_local", String(512), default=__defaults__["effect_local"]
    )
    effect_next = Column(
        "fld_effect_next", String(512), default=__defaults__["effect_next"]
    )
    effect_probability = Column(
        "fld_effect_probability", Float, default=__defaults__["effect_probability"]
    )
    hazard_rate_source = Column(
        "fld_hazard_rate_source",
        String(512),
        default=__defaults__["hazard_rate_source"],
    )
    isolation_method = Column(
        "fld_isolation_method", String(512), default=__defaults__["isolation_method"]
    )
    mission = Column("fld_mission", String(64), default=__defaults__["mission"])
    mission_phase = Column(
        "fld_mission_phase", String(64), default=__defaults__["mission_phase"]
    )
    mode_criticality = Column(
        "fld_mode_criticality", Float, default=__defaults__["mode_criticality"]
    )
    mode_hazard_rate = Column(
        "fld_mode_hazard_rate", Float, default=__defaults__["mode_hazard_rate"]
    )
    mode_op_time = Column(
        "fld_mode_op_time", Float, default=__defaults__["mode_op_time"]
    )
    mode_probability = Column(
        "fld_mode_probability", String(64), default=__defaults__["mode_probability"]
    )
    mode_ratio = Column("fld_mode_ratio", Float, default=__defaults__["mode_ratio"])
    operator_actions = Column(
        "fld_operator_actions", String, default=__defaults__["operator_actions"]
    )
    other_indications = Column(
        "fld_other_indications", String(512), default=__defaults__["other_indications"]
    )
    remarks = Column("fld_remarks", String, default=__defaults__["remarks"])
    rpn_severity = Column(
        "fld_rpn_severity", Integer, default=__defaults__["rpn_severity"]
    )
    rpn_severity_new = Column(
        "fld_rpn_severity_new", Integer, default=__defaults__["rpn_severity_new"]
    )
    severity_class = Column(
        "fld_severity_class", String(64), default=__defaults__["severity_class"]
    )
    single_point = Column(
        "fld_single_point", Integer, default=__defaults__["single_point"]
    )
    type_id = Column("fld_type_id", Integer, default=__defaults__["type_id"])

    # Define the relationships to other tables in the RAMSTK Program database.
    mechanism: relationship = relationship(
        "RAMSTKMechanismRecord",
        back_populates="mode",
        cascade="all,delete",
    )

    is_mode = True
    is_mechanism = False
    is_cause = False
    is_control = False
    is_action = False
    is_opload = False
    is_opstress = False
    is_testmethod = False

    def get_attributes(self):
        """Retrieve the current values of the RAMSTKMode data model attributes.

        :return: {revision_id, hardware_id, mode_id, critical_item,
                  description, design_provisions, detection_method,
                  effect_end, effect_local, effect_next, effect_probability,
                  hazard_rate_source, isolation_method, mission, mission_phase,
                  mode_criticality, mode_hazard_rate, mode_op_time,
                  mode_probability, mode_ratio, operator_actions,
                  other_indications, remarks, rpn_severity, rpn_severity_new,
                  severity_class, single_point, type_id} pairs.
        :rtype: dict
        """
        _attributes = {
            "revision_id": self.revision_id,
            "hardware_id": self.hardware_id,
            "mode_id": self.mode_id,
            "critical_item": self.critical_item,
            "description": self.description,
            "design_provisions": self.design_provisions,
            "detection_method": self.detection_method,
            "effect_end": self.effect_end,
            "effect_local": self.effect_local,
            "effect_next": self.effect_next,
            "effect_probability": self.effect_probability,
            "hazard_rate_source": self.hazard_rate_source,
            "isolation_method": self.isolation_method,
            "mission": self.mission,
            "mission_phase": self.mission_phase,
            "mode_criticality": self.mode_criticality,
            "mode_hazard_rate": self.mode_hazard_rate,
            "mode_op_time": self.mode_op_time,
            "mode_probability": self.mode_probability,
            "mode_ratio": self.mode_ratio,
            "operator_actions": self.operator_actions,
            "other_indications": self.other_indications,
            "remarks": self.remarks,
            "rpn_severity": self.rpn_severity,
            "rpn_severity_new": self.rpn_severity_new,
            "severity_class": self.severity_class,
            "single_point": self.single_point,
            "type_id": self.type_id,
        }

        return _attributes



# -*- coding: utf-8 -*-
#
#       ramstk.views.gtk3.desktop.py is part of The RAMSTK Project
#
# All rights reserved.
# Copyright 2007 - 2020 Doyle Rowland doyle.rowland <AT> reliaqual <DOT> com
"""The RAMSTK GTK3 basebook."""

# Standard Library Imports
import locale
from typing import List, TypeVar

# Third Party Imports
# noinspection PyPackageRequirements
from pubsub import pub

# RAMSTK Package Imports
from ramstk.configuration import RAMSTKSiteConfiguration, RAMSTKUserConfiguration
from ramstk.logger import RAMSTKLogManager
from ramstk.views.gtk3 import Gdk, GdkPixbuf, GObject, Gtk, _
from ramstk.views.gtk3.assistants import (
    CreateProject,
    ExportProject,
    ImportProject,
    OpenProject,
)
from ramstk.views.gtk3.books import RAMSTKListBook, RAMSTKModuleBook, RAMSTKWorkBook
from ramstk.views.gtk3.options import OptionsDialog
from ramstk.views.gtk3.preferences import PreferencesDialog

Tconfiguration = TypeVar(
    "Tconfiguration", RAMSTKUserConfiguration, RAMSTKSiteConfiguration
)


def destroy(__widget: Gtk.Widget, __event: Gdk.Event = None) -> None:
    """Quit the RAMSTK application.

    This function quits the RAMSTK application when the X in the upper right
    corner is pressed or if this function is called as a callback.

    :param __widget: the Gtk.Widget() that called this method.
    :param __event: the Gdk.Event() that called this method.
    :return: None
    :rtype: None
    """
    Gtk.main_quit()


class RAMSTKDesktop(Gtk.Window):
    """The base view for the RAMSTK Books.

    This is the container class for the List Book, Module Book, and Work Book.
    Attributes of the RAMSTKDesktop are:

    :cvar RAMSTK_USER_CONFIGURATION: the instance of the RAMSTK
        user configuration class.
    :type RAMSTK_USER_CONFIGURATION:
        :class:`ramstk.configuration.RAMSTKUserConfiguration`
    :cvar dict dic_books: dictionary holding a reference to each RAMSTK book.
    :cvar dict dic_tab_pos: dictionary holding the Gtk.PositionType()s for each
        of left, right, top, and bottom.
    :ivar int _n_screens: the number of monitors attached to the machine
        running RAMSTK.
    :ivar float _height: the height of the monitors attached to the machine
        running RAMSTK.
    :ivar float _width: the average width of each monitor attached to the
        machine running RAMSTK.
    :ivar menubar: the Gtk.MenuBar() for the RAMSTK ModuleBook menu.
    :type menubar: :class:`Gtk.MenuBar`
    :ivar progressbar: the Gtk.Progressbar() for displaying progress counters.
    :type progressbar: :class:`Gtk.Progressbar`
    :ivar statusbar: the Gtk.Statusbar() for displaying messages.
    :type statusbar: :class:`Gtk.Statusbar`
    :ivar toolbar: the Gtk.Toolbar() for the RAMSTK ModuleBook tools.
    :type toolbar: :class:`Gtk.Toolbar`
    """

    RAMSTK_SITE_CONFIGURATION: RAMSTKSiteConfiguration = RAMSTKSiteConfiguration()
    RAMSTK_USER_CONFIGURATION: RAMSTKUserConfiguration = RAMSTKUserConfiguration()

    def __init__(self, configuration: Tconfiguration, logger: RAMSTKLogManager) -> None:
        """Initialize an instance of the RAMSTK Book.

        :param configuration: a list containing the RAMSTK user
            configuration and RAMSTK site configuration class instances.
        :param logger: the RAMSTKLogManager class instance.
        """
        # pylint: disable=non-parent-init-called
        GObject.GObject.__init__(self)
        self.RAMSTK_USER_CONFIGURATION = configuration[0]  # type: ignore
        self.RAMSTK_SITE_CONFIGURATION = configuration[1]  # type: ignore

        # Initialize private dictionary attributes.

        # Initialize private list attributes.

        # Initialize private scalar attributes.
        self._logger = logger
        try:
            _screen: Gdk.Screen = Gdk.Screen()
            _display = _screen.get_default().get_display()
            _monitor = _display.get_monitor(0)
            self._height: float = _monitor.get_geometry().height
            self._n_screens: int = _display.get_n_monitors()
            self._width: float = _monitor.get_geometry().width
        except AttributeError:
            # When running on CI servers, there will be no monitor.  We also
            # don't need one.
            self._height = -1
            self._n_screens = 0
            self._width = -1

        # Initialize public dictionary attributes.

        # Initialize public list attributes.

        # Initialize public scalar attributes.
        self.menubar: Gtk.MenuBar = Gtk.MenuBar()
        self.progressbar: Gtk.ProgressBar = Gtk.ProgressBar()
        self.statusbar: Gtk.Statusbar = Gtk.Statusbar()
        self.toolbar: Gtk.Toolbar = Gtk.Toolbar()

        self.icoStatus: Gtk.StatusIcon = Gtk.StatusIcon()

        self.nbkListBook: RAMSTKListBook = RAMSTKListBook(
            self.RAMSTK_USER_CONFIGURATION, logger
        )
        self.nbkModuleBook: RAMSTKModuleBook = RAMSTKModuleBook(
            self.RAMSTK_USER_CONFIGURATION, logger
        )
        self.nbkWorkBook: RAMSTKWorkBook = RAMSTKWorkBook(
            self.RAMSTK_USER_CONFIGURATION, logger
        )
        self.nbkWorkBook.RAMSTK_SITE_CONFIGURATION = self.RAMSTK_SITE_CONFIGURATION

        try:
            locale.setlocale(
                locale.LC_ALL, self.RAMSTK_USER_CONFIGURATION.RAMSTK_LOCALE
            )
        except locale.Error:
            locale.setlocale(locale.LC_ALL, "")

        self.__set_properties()
        self.__make_ui()
        self.__set_callbacks()

        # Subscribe to PyPubSub messages.
        pub.subscribe(self._on_request_open, "request_open_program ")
        pub.subscribe(self._on_select, "request_set_title")
        pub.subscribe(self._do_set_status, "request_set_status")

    def _do_request_options_assistant(self, __widget: Gtk.ImageMenuItem) -> None:
        """Request the EditOptions assistant be launched.

        :param __widget: the Gtk.ImageMenuItem() that called this class.
        :return: None
        :rtype: None
        """
        _dialog = OptionsDialog(parent=self)

        # ISSUE: Make Site DB available without connecting to program DB.
        #
        # The site DB should be available without having to connect to a
        # program DB.  Currently the site DAO is a member of the
        # ProgramManager's dict of data managers.  The site DAO needs to be
        # made available out of __main__.py or the ProgramManager needs to
        # be made accessible to the RAMSTKDesktop so site options can be
        # accessed.
        # assignees: weibullguy
        # label: globalbacklog, normal
        if _dialog.do_run() == Gtk.ResponseType.OK:
            print("Need site admin or higher privileges.")

        _dialog.do_destroy()

    def _do_request_preferences_assistant(self, __widget: Gtk.ImageMenuItem) -> None:
        """Request the PreferencesDialog assistant be launched.

        :param __widget: the Gtk.ImageMenuItem() that called this class.
        :return: None
        :rtype: None
        """
        _assistant = Gtk.Window()
        _preferences = PreferencesDialog(self.RAMSTK_USER_CONFIGURATION, self._logger)

        _screen: Gdk.Screen = Gdk.Screen()
        _n_screens = _screen.get_default().get_n_monitors()
        _width = _screen.width() / _n_screens
        _height = _screen.height()

        _assistant.set_border_width(5)
        _assistant.set_default_size(_width - 450, (4 * _height / 7))
        _assistant.set_modal(True)
        _assistant.set_position(Gtk.WindowPosition.CENTER)
        _assistant.set_resizable(True)
        _assistant.set_transient_for(self)

        _assistant.add(_preferences)

        _assistant.show_all()

    def __make_menu(self) -> None:
        """Make the menu for the Module Book.

        :return: None
        :rtype: None
        """
        self.menubar.append(self.__make_menu_file())
        self.menubar.append(self.__make_menu_edit())
        self.menubar.append(self.__make_menu_tools())

        self.menubar.show_all()

    def __make_menu_edit(self) -> Gtk.MenuItem:
        """Make the Edit menu.

        :return: the Edit menu.
        :rtype: :class:`Gtk.MenuItem`
        """
        _menu = Gtk.Menu()

        _menu_items = self.__make_menu_items(["preferences"], ["_Preferences"])

        for _menu_item in _menu_items:
            _menu.append(_menu_item)

        _menu_items[0].connect("activate", self._do_request_preferences_assistant)

        _menu_item = Gtk.MenuItem(label=_("_Edit"), use_underline=True)
        _menu_item.set_submenu(_menu)

        return _menu_item

    def __make_menu_file(self) -> Gtk.MenuItem:
        """Make the File menu.

        :return: the File menu.
        :rtype: :class:`Gtk.MenuItem`
        """
        _menu = Gtk.Menu()

        _menu_items = self.__make_menu_items(
            ["new", "open", "import", "export", "save", "", "exit"],
            [
                "New _Program",
                "_Open",
                "_Import Project",
                "_Export Project",
                "_Save",
                "_Close",
                "E_xit",
            ],
        )

        for _menu_item in _menu_items:
            _menu.append(_menu_item)

        _menu_items[0].connect(
            "activate", CreateProject, self.RAMSTK_USER_CONFIGURATION, self
        )
        _menu_items[1].connect(
            "activate", OpenProject, self.RAMSTK_USER_CONFIGURATION, self
        )
        _menu_items[2].connect(
            "activate", ImportProject, self.RAMSTK_USER_CONFIGURATION, self
        )
        _menu_items[3].connect(
            "activate", ExportProject, self.RAMSTK_USER_CONFIGURATION, self
        )
        _menu_items[4].connect("activate", self._do_request_save_project)
        _menu_items[5].connect("activate", self._do_request_close_project)
        _menu_items[6].connect("activate", destroy)

        _menu_item = Gtk.MenuItem()
        _menu_item.set_label(_("_File"))
        _menu_item.set_property("use_underline", True)
        _menu_item.set_submenu(_menu)

        return _menu_item

    def __make_menu_items(
        self, icons: List[str], labels: List[str]
    ) -> List[Gtk.ImageMenuItem]:
        """Make a list of the menu items for a menu.

        :param icons: the list of icons to display on each menu item.
        :param labels: the list of label text for each menu item.
        :return: _menu_items; the list of menu items to include in the menu.
        """
        _menu_items: List[Gtk.ImageMenuItem] = []

        for _idx, _ico in enumerate(icons):
            if _ico != "":
                _icon = (
                    self.RAMSTK_USER_CONFIGURATION.RAMSTK_ICON_DIR
                    + "/16x16/"
                    + _ico
                    + ".png"
                )
            else:
                _icon = (
                    self.RAMSTK_USER_CONFIGURATION.RAMSTK_ICON_DIR
                    + "/16x16/default.png"
                )

            _menu_item = Gtk.ImageMenuItem()
            _image = Gtk.Image()
            _image.set_from_file(_icon)
            _menu_item.set_label(_(labels[_idx]))
            _menu_item.set_image(_image)
            _menu_item.set_property("use_underline", True)

            _menu_items.append(_menu_item)

        return _menu_items

    def __make_menu_tools(self) -> Gtk.MenuItem:
        """Make the Tools menu.

        :return: the Tools menu.
        :rtype: :class:`Gtk.MenuItem`
        """
        _menu = Gtk.Menu()

        _menu_items = self.__make_menu_items(["options"], ["_Options"])

        for _menu_item in _menu_items:
            _menu.append(_menu_item)

        _menu_items[0].connect("activate", self._do_request_options_assistant)

        _menu_item = Gtk.MenuItem(label=_("_Tools"), use_underline=True)
        _menu_item.set_submenu(_menu)

        return _menu_item

    def __make_toolbuttons(
        self, icons: List[str], tooltips: List[str]
    ) -> List[Gtk.ToolButton]:
        """Make the toolbuttons for a toolbar.

        :param icons: the list of icons to display on the toolbutton.
        :param tooltips: the list of tooltips to associate with each
            toolbutton.
        :return: _buttons; the list of toolbuttons to display in the toolbar.
        """
        _buttons: List[Gtk.ToolButton] = []

        for _idx, _ico in enumerate(icons):
            if _ico != "":
                _icon = (
                    self.RAMSTK_USER_CONFIGURATION.RAMSTK_ICON_DIR
                    + "/32x32/"
                    + _ico
                    + ".png"
                )
            else:
                _icon = (
                    self.RAMSTK_USER_CONFIGURATION.RAMSTK_ICON_DIR
                    + "/32x32/default.png"
                )

            _button = Gtk.ToolButton()
            _button.set_tooltip_text(tooltips[_idx])
            _image = Gtk.Image()
            _image.set_from_file(_icon)
            _button.set_icon_widget(_image)

            _buttons.append(_button)

        return _buttons

    def __make_toolbar(self) -> None:
        """Make the toolbar for the Module Book.

        :return: None
        :rtype: None
        """
        _toolbuttons = self.__make_toolbuttons(
            ["new", "open", "close", "save", "save-exit", "exit"],
            [
                _("Create a new RAMSTK Program Database."),
                _("Connect to an existing RAMSTK Program Database."),
                _("Closes the open RAMSTK Program Database."),
                _("Save the currently open RAMSTK Project Database."),
                _("Save the currently open RAMSTK Program Database then quits."),
                _("Quits without saving the currently open RAMSTK Program Database."),
            ],
        )

        for _position, _toolbutton in enumerate(_toolbuttons):
            self.toolbar.insert(_toolbutton, _position)

        self.toolbar.insert(Gtk.SeparatorToolItem(), 3)
        self.toolbar.insert(Gtk.SeparatorToolItem(), 5)

        _toolbuttons[0].connect(
            "clicked", CreateProject, self.RAMSTK_USER_CONFIGURATION, self
        )
        _toolbuttons[1].connect(
            "clicked", OpenProject, self.RAMSTK_USER_CONFIGURATION, self
        )
        _toolbuttons[2].connect("clicked", self._do_request_close_project)
        _toolbuttons[3].connect("clicked", self._do_request_save_project)
        _toolbuttons[4].connect("clicked", self._do_request_save_project, True)
        _toolbuttons[5].connect("clicked", destroy)

        self.toolbar.show_all()

    def __make_ui(self) -> None:
        """Build the user interface.

        :return: None
        :rtype: None
        """
        self.__make_menu()
        self.__make_toolbar()

        _vbox = Gtk.VBox()
        _vbox.pack_start(self.menubar, False, False, 0)
        _vbox.pack_start(self.toolbar, False, False, 0)

        _hpaned = Gtk.HPaned()
        _hpaned.pack1(self.nbkModuleBook, True, False)
        _hpaned.pack2(self.nbkListBook, True, False)

        _vpaned = Gtk.VPaned()
        _vpaned.pack1(_hpaned)
        _vpaned.pack2(self.nbkWorkBook, True, False)

        _vbox.pack_start(_vpaned, True, True, 0)
        _vbox.pack_start(self.statusbar, False, False, 0)

        self.add(_vbox)
        self.show_all()

        self.nbkModuleBook.set_current_page(0)

        self.statusbar.push(1, _("Ready"))
        self._do_set_status_icon()

    def __set_callbacks(self) -> None:
        """Set the callbacks for the RAMSTKListBook() and widgets.

        :return: None
        :rtype: None
        """
        self.connect("delete_event", destroy)
        self.connect("window_state_event", self._on_window_state_event)
        self.connect("button_press_event", self._on_button_press)

    def __set_properties(self) -> None:
        """Set properties of the RAMSTKBook and widgets.

        :return: None
        :rtype: None
        """
        self.set_border_width(5)
        self.set_position(Gtk.WindowPosition.NONE)
        self.set_resizable(True)

        _width = 15 * self._width / 16
        _height = 15 * self._height / 16
        self.resize(_width, _height)
        self.move(50, 5)

    @staticmethod
    def _do_request_close_project(__widget: Gtk.Widget) -> None:
        """Request to close the open RAMSTK Program.

        :param Gtk.Widget __widget: the Gtk.Widget() that called this method.
        :return: None
        :rtype: None
        """
        pub.sendMessage("request_close_project")

    # noinspection PyDeepBugsSwappedArgs
    def _do_request_save_project(self, widget: Gtk.Widget, end: bool = False) -> None:
        """Request to save the open RAMSTK Program.

        :param Gtk.Widget widget: the Gtk.Widget() that called this method.
        :keyword bool end: indicates whether or not to quit RAMSTK after saving
            the project.
        :return: None
        :rtype: None
        """
        _message = _("Saving Program Database {0:s}").format(
            self.RAMSTK_USER_CONFIGURATION.RAMSTK_PROG_INFO["database"]
        )
        self.statusbar.push(2, _message)

        pub.sendMessage("request_save_project")

        if end:
            destroy(widget)

    def _do_set_status(self, status: str) -> None:
        """Set the status message.

        :param status: the status message to display.
        :return: None
        :rtype: None
        """
        self.statusbar.push(1, status)

    def _do_set_status_icon(self, connected: bool = False) -> None:
        """Set the status icon in the system tay to indicate connection status.

        :param bool connected: whether or not RAMSTK is connected to a program
            database.
        :return: None
        :rtype: None
        """
        _pixbuf: GdkPixbuf.Pixbuf = GdkPixbuf.Pixbuf()

        if connected:
            _icon: str = (
                self.RAMSTK_USER_CONFIGURATION.RAMSTK_ICON_DIR
                + "/32x32/db-connected.png"
            )
            _tooltip: str = _(
                f"RAMSTK is connected to program database "
                f"{self.RAMSTK_USER_CONFIGURATION.RAMSTK_PROG_INFO['database']}."
            )
        else:
            _icon = (
                self.RAMSTK_USER_CONFIGURATION.RAMSTK_ICON_DIR
                + "/32x32/db-disconnected.png"
            )
            _tooltip = _("RAMSTK is not currently connected to a project database.")

        self.icoStatus.set_from_pixbuf(_pixbuf.new_from_file_at_size(_icon, 22, 22))
        self.icoStatus.set_tooltip_markup(_tooltip)

    def _on_button_press(self, __book: object, event: Gdk.EventButton) -> None:
        """Handle mouse clicks on the RAMSTKBook.

        :param __book: the RAMSTKBook that was 'clicked'.
        :param event: the Gdk.Event() that called this method (the
                      important attribute is which mouse button was clicked).

                      * 1 = left
                      * 2 = scrollwheel
                      * 3 = right
                      * 4 = forward
                      * 5 = backward
                      * 8 =
                      * 9 =

        :type event: :class:`Gdk.Event`
        :return: None
        :rtype: None
        """
        if event.button == 1:
            self.present()

    def _on_request_open(self) -> None:
        """Set the status bar and update the progress bar.

        :return: None
        :rtype: None
        """
        _message = _(
            f"Opening Program Database "
            f"{self.RAMSTK_USER_CONFIGURATION.RAMSTK_PROG_INFO['database']}"
        )

        # noinspection PyDeepBugsSwappedArgs
        self.statusbar.push(1, _message)
        self.set_title(
            _(
                f"RAMSTK - Analyzing "
                f"{self.RAMSTK_USER_CONFIGURATION.RAMSTK_PROG_INFO['database']}"
            )
        )

    def _on_select(self, title: str) -> None:
        """Respond to load the Work View Gtk.Notebook() widgets.

        This method handles the results of the an individual module's
        _on_select() method.  It sets the title of the RAMSTK Work Book and
        raises an error dialog if needed.

        :return: None
        :rtype: None
        """
        try:
            self.set_title(title)
        except AttributeError as _error:
            self.RAMSTK_LOGGER.do_log_exception(__name__, _error)
            self.do_raise_dialog(severity="warning", user_msg=_error)

    @staticmethod
    def _on_window_state_event(window: Gtk.Window, event: Gdk.EventWindowState) -> None:
        """Iconify or deiconify the desktop.

        :return: None
        :rtype: None
        """
        if event.new_window_state == Gdk.WindowState.ICONIFIED:
            window.iconify()
        elif event.new_window_state == 0:
            window.deiconify()
        elif event.new_window_state == Gdk.WindowState.MAXIMIZED:
            window.maximize()



# -*- coding: utf-8 -*-
#
#       ramstk.views.gtk3.similar_item.view.py is part of the RAMSTK Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""GTK3 Similar Item Views."""

# Standard Library Imports
from typing import Any, Dict

# Third Party Imports
from pubsub import pub

# RAMSTK Package Imports
from ramstk.configuration import RAMSTKUserConfiguration
from ramstk.logger import RAMSTKLogManager
from ramstk.views.gtk3 import Gtk, _
from ramstk.views.gtk3.assistants import EditFunction
from ramstk.views.gtk3.widgets import RAMSTKPanel, RAMSTKWorkView

# RAMSTK Local Imports
from . import SimilarItemMethodPanel, SimilarItemTreePanel


class SimilarItemWorkView(RAMSTKWorkView):
    """Display Similar Item attribute data in the Work Book.

    The WorkView displays all the attributes for the Similar Item Analysis. The
    attributes of a SimilarItem Work View are:

    :cvar dict _dic_quality: the quality levels and associated index to use in
        a Topic 633 analysis.
    :cvar dict _dic_environment: the environments and associated index to use
        in a Topic 633 analysis.
    :cvar str _tag: the name of the module.

    :ivar dict _dic_hardware: dict to hold information from the Hardware
        module.
    :ivar list _lst_callbacks: the list of callback methods for the view's
        toolbar buttons and pop-up menu.  The methods are listed in the order
        they appear on the toolbar and pop-up menu.
    :ivar list _lst_icons: the list of icons for the view's toolbar buttons
        and pop-up menu.  The icons are listed in the order they appear on the
        toolbar and pop-up menu.
    :ivar list _lst_mnu_labels: the list of labels for the view's pop-up
        menu.  The labels are listed in the order they appear in the menu.
    :ivar list _lst_tooltips: the list of tooltips for the view's
        toolbar buttons and pop-up menu.  The tooltips are listed in the
        order they appear on the toolbar or pop-up menu.
    :ivar int _hardware_id: the Hardware ID of the selected Similar Item.
    :ivar int _method_id: the ID of the similar item method to use.
    """

    # Define private dict class attributes.

    # Define private list class attributes.

    # Define private scalar class attributes.
    _tag: str = "similar_item"
    _tablabel: str = _("Similar Item")
    _tabtooltip: str = _(
        "Displays the Similar Item analysis for the selected hardware item."
    )

    # Define public dictionary class attributes.

    # Define public dictionary list attributes.

    # Define public dictionary scalar attributes.

    def __init__(
        self, configuration: RAMSTKUserConfiguration, logger: RAMSTKLogManager
    ) -> None:
        """Initialize the Hardware Work View general data page.

        :param configuration: the RAMSTKUserConfiguration class instance.
        :type configuration:
            :class:`ramstk.configuration.RAMSTKUserConfiguration`
        :param logger: the RAMSTKLogManager class instance.
        :type logger: :class:`ramstk.logger.RAMSTKLogManager`
        """
        super().__init__(configuration, logger)

        # Initialize private dictionary attributes.

        # Initialize private list attributes.
        self._lst_callbacks.insert(0, self._do_request_edit_function)
        self._lst_callbacks.insert(1, self._do_request_rollup)
        self._lst_callbacks.insert(2, self._do_request_calculate)
        self._lst_icons.insert(0, "edit")
        self._lst_icons.insert(1, "rollup")
        self._lst_icons.insert(2, "calculate")
        self._lst_mnu_labels = [
            _("Edit Function"),
            _("Roll-Up Descriptions"),
            _("Calculate Similar Item"),
            _("Save"),
            _("Save All"),
        ]
        self._lst_tooltips = [
            _("Edit the similar item analysis functions."),
            _("Roll up descriptions to next higher level assembly."),
            _("Calculate the similar item analysis."),
            _("Save changes to the selected similar item analysis line item."),
            _("Save changes to all similar item analysis line items."),
        ]

        # Initialize private scalar attributes.
        self._pnlMethod: RAMSTKPanel = SimilarItemMethodPanel()
        self._pnlPanel: RAMSTKPanel = SimilarItemTreePanel()

        # Initialize public dictionary attributes.

        # Initialize public list attributes.

        # Initialize public scalar attributes.

        self.__make_ui()

        # Subscribe to PyPubSub messages.
        pub.subscribe(
            super().do_set_cursor_active,
            "succeed_roll_up_change_descriptions",
        )

        pub.subscribe(
            self._do_set_record_id,
            "selected_hardware",
        )

    def _do_set_record_id(self, attributes: Dict[str, Any]) -> None:
        """Set the allocation's record ID.

        :param attributes: the attributes dict for the selected allocation
            record.
        :return: None
        :rtype: None
        """
        self.dic_pkeys["parent_id"] = attributes["parent_id"]
        self.dic_pkeys["record_id"] = attributes["hardware_id"]

    def _do_request_calculate(self, __button: Gtk.ToolButton) -> None:
        """Request to iteratively calculate the Similar Item metrics.

        :param __button: the Gtk.ToolButton() that called this method.
        :return: None
        :rtype: None
        """
        _model = self._pnlPanel.tvwTreeView.get_model()
        _row = _model.get_iter_first()

        # Iterate through the assemblies and calculate the Similar Item hazard
        # intensities.
        super().do_set_cursor_busy()
        _node_ids = []
        while _row is not None:
            _node_ids.append(_model.get_value(_row, 1))
            _row = _model.iter_next(_row)

        for _node_id in _node_ids:
            pub.sendMessage("request_calculate_similar_item", node_id=_node_id)

    def _do_request_edit_function(self, __button: Gtk.ToolButton) -> None:
        """Request to edit the Similar Item analysis user-defined functions.

        :param __button: the Gtk.ToolButton() that called this method.
        :return: None
        :rtype: None
        """
        (
            _model,
            _row,
        ) = self._pnlPanel.tvwTreeView.get_selection().get_selected()  # noqa

        _dialog = EditFunction(
            self._pnlPanel.tvwTreeView,
            dlgparent=self.get_parent().get_parent().get_parent().get_parent(),
        )

        if _dialog.do_run() == Gtk.ResponseType.OK:
            _functions = _dialog.do_set_functions(self._pnlPanel.tvwTreeView)
            if _dialog.chkApplyAll.get_active():
                _row = _model.get_iter_first()
                while _row is not None:
                    self._pnlPanel.do_refresh_functions(_row, _functions)
                    _row = _model.iter_next(_row)
            else:
                self._pnlPanel.do_refresh_functions(_row, _functions)

        _dialog.do_destroy()

    def _do_request_rollup(self, __button: Gtk.ToolButton) -> None:
        """Request to roll-up the Similar Item change descriptions.

        :param __button: the Gtk.ToolButton() that called this method.
        :return: None
        :rtype: None
        """
        super().do_set_cursor_busy()
        pub.sendMessage(
            "request_roll_up_change_descriptions",
            node=self._pnlPanel.tree.get_node(self.dic_pkeys["record_id"]),
        )

    def __make_ui(self) -> None:
        """Build the user interface for the Similar Item tab.

        :return: None
        :rtype: None
        """
        _hpaned: Gtk.HPaned = super().do_make_layout_lr()

        super().do_embed_treeview_panel()
        self._pnlPanel.do_load_comboboxes()

        self.remove(self.get_children()[-1])
        _hpaned.pack1(self._pnlMethod, True, True)
        _hpaned.pack2(self._pnlPanel, True, True)

        self.show_all()



# -*- coding: utf-8 -*-
#
#       ramstk.analyses.statistics.bounds.py is part of The RAMSTK Project
#
# All rights reserved.
# Copyright 2007 - 2017 Doyle Rowland doyle.rowland <AT> reliaqual <DOT> com
"""Functions for performing calculations associated with statistical bounds."""

# Standard Library Imports
import inspect
from typing import Callable, List, Tuple

# Third Party Imports
import numpy as np
from scipy import misc
from scipy.stats import norm


def do_calculate_beta_bounds(
    minimum: float, likely: float, maximum: float, alpha: float
) -> Tuple[float, float, float, float]:
    """Calculate the mean, standard error, and bounds of the beta distribution.

    These are the project management estimators, not exact calculations.

    :param minimum: the minimum expected value.
    :param likely: most likely value.
    :param maximum: the maximum expected value.
    :param alpha: the desired confidence level.
    :return: _meanll, _mean, _meanul, _sd; the calculated mean, bounds, and
                                           standard error.
    :rtype: tuple of floats
    """
    if alpha > 1.0:
        _z_norm = norm.ppf(1.0 - ((1.0 - alpha / 100.0) / 2.0))
    else:
        _z_norm = norm.ppf(1.0 - ((1.0 - alpha) / 2.0))

    _mean = (minimum + 4.0 * likely + maximum) / 6.0
    _sd = (maximum - minimum) / 6.0

    _meanll = _mean - _z_norm * _sd
    _meanul = _mean + _z_norm * _sd

    return _meanll, _mean, _meanul, _sd


def do_calculate_fisher_information(
    model: Callable, p0: List[float], data: np.ndarray, noise=1.0
) -> np.ndarray:
    """Calculate the Fisher information matrix for model.

    Sampled on grid X with parameters p0.  Assumes samples are not correlated
    and have equal variance noise^2.  This function is called, for example:

    >>> _data = np.array([[0.0, 1.585, 1, 1, 1.585],
                          [0.0, 1.978, 1, 1, 1.978],
                          ...
                          [0.0, 350.577, 1, 1, 350.577],
                          [0.0, 351.347, 1, 1, 351.347]])
    >>> _p0 = [0.010623498434893014, 0.0]
    >>> _model = log_pdf(data, theta, loc = 0.0)
    >>> do_calculate_fisher_information(_model, _p0, _data)
        array([[8.67337390e+05, 9.89376513e+01],
               [9.89376513e+01, 1.12858721e-02]])

    :param model: the model function, f(x, ...). This function must take the
        data set as the first argument.  The remaining arguments of the
        function should be the scale, shape, and location parameters.
    :param p0: point in parameter space where Fisher information matrix is
        evaluated.  Passed as a list in the same order as the parameter
        arguments to the model.  See the example above.
    :param data: the data set to use for calculating the information matrix.
    :param noise: squared variance of the noise in data.
    :returns: _fisher; the Fisher information matrix.
    :rtype: ndarray
    """
    _labels = inspect.getfullargspec(model)[0][1:]
    _p0dict = dict(zip(_labels, p0))

    _D = np.zeros((len(p0), data.size))

    for i, argname in enumerate(_labels):
        # pylint: disable=cell-var-from-loop
        _D[i, :] = [
            misc.derivative(
                lambda p: model(_record, **dict(_p0dict, **{argname: p})),
                _p0dict[argname],
                dx=1.0e-6,
            )
            for _record in data
        ]

    _fisher = 1.0 / noise ** 2 * np.einsum("mk, nk", _D, _D)

    return _fisher



# -*- coding: utf-8 -*-
#
#       ramstk.analyses.prediction.Filter.py is part of the RAMSTK Project
#
# All rights reserved.
# Copyright 2007 - 2017 Doyle Rowland doyle.rowland <AT> reliaqual <DOT> com
"""Filter MIL-HDBK-217F Constants and Calculations Module."""

# Standard Library Imports
from typing import Any, Dict

PART_COUNT_LAMBDA_B = {
    1: [
        0.022,
        0.044,
        0.13,
        0.088,
        0.20,
        0.15,
        0.20,
        0.24,
        0.29,
        0.24,
        0.018,
        0.15,
        0.33,
        2.6,
    ],
    2: [0.12, 0.24, 0.72, 0.48, 1.1, 0.84, 1.1, 1.3, 1.6, 1.3, 0.096, 0.84, 1.8, 1.4],
    3: [0.27, 0.54, 1.6, 1.1, 2.4, 1.9, 2.4, 3.0, 3.5, 3.0, 0.22, 1.9, 4.1, 32.0],
}
PART_STRESS_LAMBDA_B = {1: 0.022, 2: 0.12, 3: 0.12, 4: 0.27}
PI_E = [1.0, 2.0, 6.0, 4.0, 9.0, 7.0, 9.0, 11.0, 13.0, 11.0, 0.8, 7.0, 15.0, 120.0]
PI_Q = [1.0, 2.9]


def calculate_part_count(**attributes: Dict[str, int]) -> float:
    """Wrap get_part_count_lambda_b().

    This wrapper allows us to pass an attributes dict from a generic parts
    count function.

    :param attributes: the attributes for the filter being calculated.
    :return: _base_hr; the parts count base hazard rate.
    :rtype: float
    """
    return get_part_count_lambda_b(
        attributes["type_id"], attributes["environment_active_id"]  # type: ignore
    )  # type: ignore


def calculate_part_stress(**attributes: Dict[str, Any]) -> Dict[str, Any]:
    """Calculate the part stress active hazard rate for a filter.

    :param attributes: the attributes for the filter being calculated.
    :return: attributes; the keyword argument (hardware attribute)
             dictionary with updated values.
    :rtype: dict
    :raise: KeyError if an unknown type ID is passed.
    """
    attributes["lambda_b"] = PART_STRESS_LAMBDA_B[  # type: ignore
        attributes["type_id"]  # type: ignore
    ]  # type: ignore

    attributes["hazard_rate_active"] = (
        attributes["lambda_b"] * attributes["piQ"] * attributes["piE"]  # type: ignore
    )

    return attributes


def get_part_count_lambda_b(type_id: int, environment_active_id: int) -> float:
    """Retrievee the part count base hazard rate for a filter.

    :param type_id: the filter type identifer.
    :param environment_active_id: the active environment identifier.
    :return: _base_hr; the part count base hazard rate for the active
        environment.
    :rtype: float
    :raise: IndexError if an unknown active environment ID is passed.
    :raise: KeyError if an unknown type ID is passed.
    """
    return PART_COUNT_LAMBDA_B[type_id][environment_active_id - 1]



# pylint: disable=unused-import
# -*- coding: utf-8 -*-
#
#       ramstk.views.gtk3.widgets.__init__.py is part of The RAMSTK Project
#
# All rights reserved.
# Copyright 2019 Doyle Rowland doyle.rowland <AT> reliaqual <DOT> com
"""The RAMSTK GTK3 widgets package."""

# RAMSTK Local Imports
from .basebook import RAMSTKBaseBook
from .baseview import RAMSTKBaseView, RAMSTKListView, RAMSTKModuleView, RAMSTKWorkView
from .button import (
    RAMSTKButton,
    RAMSTKCheckButton,
    RAMSTKFileChooserButton,
    RAMSTKOptionButton,
    RAMSTKSpinButton,
    do_make_buttonbox,
)
from .combo import RAMSTKComboBox
from .dialog import (
    RAMSTKDatabaseSelect,
    RAMSTKDateSelect,
    RAMSTKDialog,
    RAMSTKFileChooser,
    RAMSTKMessageDialog,
)
from .entry import RAMSTKEntry, RAMSTKTextView
from .frame import RAMSTKFrame
from .label import RAMSTKLabel, do_make_label_group
from .panel import RAMSTKFixedPanel, RAMSTKPanel, RAMSTKPlotPanel, RAMSTKTreePanel
from .plot import RAMSTKPlot
from .scrolledwindow import RAMSTKScrolledWindow
from .treeview import RAMSTKTreeView
from .widget import RAMSTKWidget



# -*- coding: utf-8 -*-
#
#       ramstk.models.commondb.RAMSTKMeasurement.py is part of The RAMSTK
#       Project
#
# All rights reserved.
# Copyright 2007 - 2017 Doyle Rowland doyle.rowland <AT> reliaqual <DOT> com
"""RAMSTKMeasurement Table Module."""

# Third Party Imports
from sqlalchemy import Column, Integer, String

# RAMSTK Package Imports
from ramstk.db import RAMSTK_BASE
from ramstk.models import RAMSTKBaseRecord


class RAMSTKMeasurement(RAMSTK_BASE, RAMSTKBaseRecord):
    """Class to represent ramstk_measurement in the RAMSTK Common database."""

    __defaults__ = {
        "code": "Measurement Code",
        "description": "Measurement Description",
        "measurement_type": "unknown",
    }
    __tablename__ = "ramstk_measurement"
    __table_args__ = {"extend_existing": True}

    measurement_id = Column(
        "fld_measurement_id",
        Integer,
        primary_key=True,
        autoincrement=True,
        nullable=False,
    )
    code = Column("fld_code", String(128), default=__defaults__["code"])
    description = Column(
        "fld_description", String(512), default=__defaults__["description"]
    )
    measurement_type = Column(
        "fld_type", String(128), default=__defaults__["measurement_type"]
    )

    def get_attributes(self):
        """Retrieve current values of RAMSTKMeasurement data model attributes.

        :return: {measurement_id, description} pairs.
        :rtype: dict
        """
        _attributes = {
            "measurement_id": self.measurement_id,
            "code": self.code,
            "description": self.description,
            "measurement_type": self.measurement_type,
        }

        return _attributes



# Standard Library Imports
from typing import Any, Callable, Dict, List

# Third Party Imports
from treelib import Tree

# RAMSTK Package Imports
from ramstk.models import RAMSTKBaseView as RAMSTKBaseView

class RAMSTKHardwareBoMView(RAMSTKBaseView):
    _tag: str = ...
    _dic_load_functions: Dict[str, Callable] = ...
    _dic_stress_limits: Dict[int, List[float]] = ...
    _dic_trees: Dict[str, Tree] = ...
    _lst_modules: List[str] = ...
    _hr_multiplier: float = ...
    def __init__(self, **kwargs: Dict[Any, Any]) -> None: ...
    def do_calculate_cost(self, node_id: int) -> None: ...
    def do_calculate_hardware(self, node_id: int) -> None: ...
    def do_calculate_part_count(self, node_id: int) -> None: ...
    def do_calculate_power_dissipation(self, node_id: int) -> float: ...
    def do_make_composite_ref_des(self, node_id: int = ...) -> None: ...
    def _do_load_hardware(self) -> None: ...
    def _do_load_design_electric(self) -> None: ...
    def _do_load_design_mechanic(self) -> None: ...
    def _do_load_milhdbk217f(self) -> None: ...
    def _do_load_nswc(self) -> None: ...
    def _do_load_reliability(self) -> None: ...



# Standard Library Imports
from typing import Any

# Third Party Imports
from treelib import Tree as Tree

# RAMSTK Package Imports
from ramstk.configuration import RAMSTKUserConfiguration as RAMSTKUserConfiguration
from ramstk.logger import RAMSTKLogManager as RAMSTKLogManager
from ramstk.views.gtk3 import Gtk as Gtk
from ramstk.views.gtk3.function import FunctionModuleView as FunctionModuleView
from ramstk.views.gtk3.revision import RevisionModuleView as RevisionModuleView
from ramstk.views.gtk3.widgets import RAMSTKBaseBook as RAMSTKBaseBook

class RAMSTKModuleBook(RAMSTKBaseBook):
    _dic_module_views: Any
    icoStatus: Any
    def __init__(
        self, configuration: RAMSTKUserConfiguration, logger: RAMSTKLogManager
    ) -> None: ...
    def __make_ui(self) -> None: ...
    def __set_callbacks(self) -> None: ...
    def _on_close(self) -> None: ...
    def _on_open(self, tree: Tree) -> None: ...
    def _on_switch_page(
        self, __notebook: Gtk.Notebook, __page: Gtk.Widget, page_num: int
    ) -> None: ...
